import torch.nn as nn


class ConvBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        dropout=False,
        norm=None,
        residual=True,
        activation='leakyrelu',
        in_place_activation=True,
        transpose=False,
        reflectpad=True,
    ):

        super(ConvBlock, self).__init__()
        self.dropout = dropout
        self.residual = residual
        self.activation = activation
        self.transpose = transpose
        self.reflectpad = reflectpad

        if self.dropout:
            self.dropout1 = nn.Dropout2d(p=0.05)
            self.dropout2 = nn.Dropout2d(p=0.05)

        self.norm1 = None
        self.norm2 = None
        if norm is not None:
            if norm == 'batch':
                self.norm1 = nn.BatchNorm2d(out_channels)
                self.norm2 = nn.BatchNorm2d(out_channels)
            elif norm == 'instance':
                self.norm1 = nn.InstanceNorm2d(out_channels, affine=True)
                self.norm2 = nn.InstanceNorm2d(out_channels, affine=True)

        if self.transpose:
            self.conv1 = nn.ConvTranspose2d(
                in_channels,
                out_channels,
                kernel_size=3,
                padding=0 if self.reflectpad else 1,
            )
            self.conv2 = nn.ConvTranspose2d(
                out_channels,
                out_channels,
                kernel_size=3,
                padding=0 if self.reflectpad else 1,
            )
        else:
            self.conv1 = nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=3,
                padding=0 if self.reflectpad else 1,
            )
            self.conv2 = nn.Conv2d(
                out_channels,
                out_channels,
                kernel_size=3,
                padding=0 if self.reflectpad else 1,
            )

        if self.activation == 'relu':
            self.actfun1 = nn.ReLU(inplace=in_place_activation)
            self.actfun2 = nn.ReLU(inplace=in_place_activation)
        elif self.activation == 'leakyrelu':
            self.actfun1 = nn.LeakyReLU(inplace=in_place_activation)
            self.actfun2 = nn.LeakyReLU(inplace=in_place_activation)
        elif self.activation == 'elu':
            self.actfun1 = nn.ELU(inplace=in_place_activation)
            self.actfun2 = nn.ELU(inplace=in_place_activation)
        elif self.activation == 'selu':
            self.actfun1 = nn.SELU(inplace=in_place_activation)
            self.actfun2 = nn.SELU(inplace=in_place_activation)

        if self.reflectpad:
            self.rpad1 = nn.ReflectionPad2d(1)
            self.rpad2 = nn.ReflectionPad2d(1)

    def forward(self, x):

        ox = x

        if self.reflectpad:
            x = self.rpad1(x)

        x = self.conv1(x)

        if self.dropout:
            x = self.dropout1(x)

        x = self.actfun1(x)

        if self.norm1:
            x = self.norm1(x)

        if self.reflectpad:
            x = self.rpad2(x)

        x = self.conv2(x)

        if self.dropout:
            x = self.dropout2(x)

        if self.residual:
            x[:, 0 : min(ox.shape[1], x.shape[1]), :, :] += ox[
                :, 0 : min(ox.shape[1], x.shape[1]), :, :
            ]

        x = self.actfun2(x)

        if self.norm2:
            x = self.norm2(x)

        # print("shapes: x:%s ox:%s " % (x.shape,ox.shape))

        return x



import numpy

from aydin.io.datasets import normalise, pollen, add_noise
from aydin.it.transforms.deskew import DeskewTransform


def demo_deskew():
    shifts = tuple((3 * i, 0) for i in range(100))

    image = normalise(pollen())[0:256, 0:256]
    array = numpy.stack(
        [add_noise(numpy.roll(image, shift=shift, axis=(0, 1))) for shift in shifts]
    )

    sd = DeskewTransform(delta=-3, z_axis=0, skew_axis=1)

    ds_array = sd.preprocess(array)
    s_array = sd.postprocess(ds_array)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(array, name='array')
        viewer.add_image(ds_array, name='ds_array')
        viewer.add_image(s_array, name='s_array')


demo_deskew()



import numpy

from aydin.io.datasets import camera
from aydin.it.transforms.histogram import HistogramEqualisationTransform


def demo_histogram():
    image = camera()

    ht = HistogramEqualisationTransform()

    preprocessed = ht.preprocess(image)
    postprocessed = ht.postprocess(preprocessed)

    import napari

    viewer = napari.Viewer()
    viewer.add_image(image, name='image')
    viewer.add_image(preprocessed, name='preprocessed')
    viewer.add_image(postprocessed, name='postprocessed')
    napari.run()

    assert postprocessed.shape == image.shape
    assert numpy.abs(postprocessed - image).mean() < 1e-5


demo_histogram()



from qtpy.QtWidgets import QTreeWidgetItemIterator


def iter_tree_widget(root):
    iterator = QTreeWidgetItemIterator(root)
    while True:
        item = iterator.value()
        if item is not None:
            yield item
            iterator += 1
        else:
            break



from typing import Sequence
import numpy


def extend_nd(available_dims: Sequence[int] = (2,)):
    """
    Decorator that extends to nD a denoising function that is limited to images
    of a few dimensions (e.g. 2 or 3).

    Parameters
    ----------
    available_dims: Tuple[int]

    Returns
    -------

    """

    # Actual decorator that varies given the parameters
    def decorator(function):
        def wrapper(*args, **kwargs):

            # The first argument must be the image:
            image = args[0]

            # The dimension of the image:
            ndim = image.ndim

            if ndim in available_dims:
                # If we can denoise directly, let's do it:
                return function(*args, **kwargs)
            else:

                # What are the smallest available image dimensions that we can denoise?
                smallest_dim = min(available_dims)

                if ndim < smallest_dim:
                    # In this case we are trying to denoise an image with less dimensions that what the function can do,
                    # So we need to add extra dimensions:

                    extended_image = image
                    for _ in range(smallest_dim - ndim):
                        extended_image = extended_image[numpy.newaxis, ...]

                    # we apply the denoising function:
                    denoised = function(extended_image, *args[1:], **kwargs)

                else:
                    # In this case we try to denoise an image with more dimensions that what the function can do,
                    # so we have to denoise hyperplane by hyperplane.

                    # In this case, this is the largest available dimension below the image dimension:
                    largest_dim = max((n for n in available_dims if n < ndim))

                    # Let's use order the image by increasing dimension length,
                    # this way we will denoise the largest sub images first:
                    # The corresponding permutation is:
                    permutation = tuple(numpy.argsort(image.shape))

                    # let's permutate the array dimensions accordingly:
                    reshaped_image = numpy.transpose(image, axes=permutation)

                    # Let's collapse the first ndim-largest_dim dimensions:
                    collapsed_image = numpy.reshape(
                        reshaped_image,
                        newshape=(-1,) + reshaped_image.shape[ndim - largest_dim :],
                    )

                    # We denoise each such image:
                    collapsed_denoised = numpy.stack(
                        tuple(function(i, *args[1:], **kwargs) for i in collapsed_image)
                    )

                    # Give it back its original shape:
                    reshaped_denoised = numpy.reshape(
                        collapsed_denoised, newshape=reshaped_image.shape
                    )
                    denoised = numpy.transpose(
                        reshaped_denoised, axes=tuple(numpy.argsort(permutation))
                    )

                return denoised

        return wrapper

    return decorator



# flake8: noqa

import numpy
import numpy as np
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import normalise, add_noise, newyork
from aydin.it.classic_denoisers.gaussian import denoise_gaussian
from aydin.it.classic_denoisers.spectral import denoise_spectral
from aydin.util.j_invariance.j_invariant_classic import calibrate_denoiser_classic
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart
from aydin.util.crop.rep_crop import representative_crop
from aydin.util.log.log import Log


def demo_j_invariant_classic(image, display=True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True
    Log.set_log_max_depth(5)

    image = normalise(image.astype(np.float32))
    noisy = add_noise(image)

    # obtain representative crop, to speed things up...
    crop = representative_crop(noisy)

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {'sigma': (0.0, 1.0), 'truncate': [4, 1]}

    # Calibrate denoiser
    best_parameters = calibrate_denoiser_classic(
        crop,
        denoise_gaussian,
        denoise_parameters=parameter_ranges,
        display_images=False,
    )

    denoised = denoise_gaussian(noisy, **best_parameters)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)
    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)
    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)
    print("         noisy   :", psnr_noisy, ssim_noisy)
    print("spectral denoised:", psnr_denoised, ssim_denoised)

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='image')
            viewer.add_image(noisy, name='noisy')
            viewer.add_image(denoised, name='denoised')

    assert ssim_denoised > 0.64

    return ssim_denoised


if __name__ == "__main__":
    demo_j_invariant_classic(newyork())



from functools import partial
from typing import Sequence, Union, Optional, Tuple

import numpy
from numba import jit
from numpy.fft import fftshift, ifftshift
from scipy.fft import fftn, ifftn

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart

__fastmath = {'contract', 'afn', 'reassoc'}
__error_model = 'numpy'


def calibrate_denoise_butterworth(
    image,
    isotropic: bool = False,
    axes: Optional[Tuple[int, ...]] = None,
    max_padding: int = 32,
    min_freq: float = 0.001,
    max_freq: float = 1.0,
    min_order: float = 0.5,
    max_order: float = 6.0,
    crop_size_in_voxels: Optional[int] = 128000,
    max_num_evaluations: int = 512,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the Butterworth denoiser for the given image and returns the optimal
    parameters obtained using the N2S loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate Sobolev denoiser for.

    isotropic: bool
        When True, the filtering is isotropic
        i.e. all frequency cutoffs are the same along all axis,
        but when false, the frequency cutoffs are different for different axis.
        Anisotropic filtering is usefull for example for 3D microscopy images
        that have a different resolution along z than along x and y, or for nD+t
        images that have a very different correlation structure along time than
        along space.

    axes: Optional[Tuple[int,...]]
        Axes over which to apply low-pass filtering.
        (advanced)

    max_padding: int
        Maximum amount of padding to be added to avoid edge effects.
        (advanced)

    min_freq: float
        Minimum cutoff frequency to use for calibration. Must be within [0,1],
        typically close to zero.
        (advanced)

    max_freq: float
        Maximum cutoff frequency to use for calibration. Must be within [0,1],
        typically close to one.
        (advanced)

    max_order: float
        Maximal order for the Butterworth filter to use for calibration.
        (advanced)

    max_order: float
        Minimal order for the Butterworth filter to use for calibration.
        (advanced)

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate denoiser.
        (advanced)

    max_num_evaluations: int
        Maximum number of evaluations for finding the optimal parameters.
        (advanced)

    display_images: bool
        When True the denoised images encountered during optimisation are shown.
        (advanced)

    other_fixed_parameters: dict
        Any other fixed parameters. (advanced)

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # ranges:
    freq_cutoff_range = (min_freq, max_freq)
    order_range = (min_order, max_order)

    # Combine fixed parameters:
    other_fixed_parameters = other_fixed_parameters | {
        'max_padding': max_padding,
        'axes': axes,
    }

    if isotropic:
        # Partial function:
        _denoise_butterworth = partial(
            denoise_butterworth, **(other_fixed_parameters | {'multi_core': False})
        )

        # Parameters to test when calibrating the denoising algorithm
        parameter_ranges = {'freq_cutoff': freq_cutoff_range, 'order': order_range}
    else:
        # Partial function with parameter impedance match:
        def _denoise_butterworth(*args, **kwargs):
            _freq_cutoff = tuple(
                kwargs.pop(f'freq_cutoff_{i}') for i in range(image.ndim)
            )
            return denoise_butterworth(
                *args,
                freq_cutoff=_freq_cutoff,
                **(kwargs | other_fixed_parameters | {'multi_core': False}),
            )

        # Parameters to test when calibrating the denoising algorithm
        parameter_ranges = {
            f'freq_cutoff_{i}': freq_cutoff_range for i in range(image.ndim)
        } | {'order': order_range}

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_smart(
            crop,
            _denoise_butterworth,
            denoise_parameters=parameter_ranges,
            max_num_evaluations=max_num_evaluations,
            display_images=display_images,
        )
        | other_fixed_parameters
    )

    if not isotropic:
        # We need to adjust a bit the type of parameters passed to the denoising function:
        freq_cutoff = tuple(
            best_parameters.pop(f'freq_cutoff_{i}') for i in range(image.ndim)
        )
        best_parameters |= {'freq_cutoff': freq_cutoff}

    # Memory needed:
    memory_needed = 6 * image.nbytes  # complex numbers and more

    return denoise_butterworth, best_parameters, memory_needed


def denoise_butterworth(
    image,
    axes: Optional[Tuple[int, ...]] = None,
    freq_cutoff: Union[float, Sequence[float]] = 0.5,
    order: float = 1,
    max_padding: int = 32,
    multi_core: bool = True,
):
    """
    Denoises the given image by applying a configurable <a
    href="https://en.wikipedia.org/wiki/Butterworth_filter">Butterworth
    lowpass filter</a>. Remarkably good when your signal
    does not have high-frequencies beyond a certain cutoff frequency.
    This is probably the first algorithm that should be tried of all
    currently available in Aydin. It is actually quite impressive how
    well this performs in practice. If the signal in your images is
    band-limited as is often the case for microscopy images, this
    denoiser will work great.
    \n\n
    Note: We recommend applying a variance stabilisation transform
    to improve results for images with non-Gaussian noise.

    Parameters
    ----------
    image: ArrayLike
        Image to be denoised

    axes: Optional[Tuple[int,...]]
        Axes over which to apply lowpass filtering.

    freq_cutoff: float
        Cutoff frequency, must be within [0, 1]

    order: float
        Filter order, typically an integer above 1.

    max_padding: int
        Maximum amount of padding to be added to avoid edge effects.

    multi_core: bool
        By default we use as many cores as possible, in some cases, for small
        (test) images, it might be faster to run on a single core instead of
        starting the whole parallelization machinery.


    Returns
    -------
    Denoised image

    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # Normalise freq_cutoff argument to tuple:
    if type(freq_cutoff) is not tuple:
        freq_cutoff = tuple((freq_cutoff,) * image.ndim)

    # Number of workers:
    workers = -1 if multi_core else 1

    # Default axes:
    if axes is None:
        axes = tuple(range(image.ndim))

    # Selected axes:
    selected_axes = tuple((a in axes) for a in range(image.ndim))

    # First we need to pad the image.
    # By how much? this depends on how much low filtering we need to do:
    pad_width = tuple(
        ((_apw(fc, max_padding), _apw(fc, max_padding)) if sa else (0, 0))
        for sa, fc in zip(selected_axes, freq_cutoff)
    )

    # pad image:
    image = numpy.pad(image, pad_width=pad_width, mode='reflect')

    # Move to frequency space:
    image_f = fftn(image, workers=workers, axes=axes)

    # Center frequencies:
    image_f = fftshift(image_f, axes=axes)

    # Compute squared distance image:
    f = _compute_distance_image(freq_cutoff, image, selected_axes)

    # Chose filter implementation:
    filter = jit(nopython=True, parallel=multi_core)(_filter)

    # Apply filter:
    image_f = filter(image_f, f, order)

    # Shift back:
    image_f = ifftshift(image_f, axes=axes)

    # Back in real space:
    denoised = numpy.real(ifftn(image_f, workers=workers, axes=axes))

    # Crop to remove padding:
    denoised = denoised[tuple(slice(u, -v) for u, v in pad_width)]

    return denoised


def _compute_distance_image(freq_cutoff, image, selected_axes):
    f = numpy.zeros_like(image, dtype=numpy.float32)
    axis_grid = tuple(
        (numpy.linspace(-1, 1, s) if sa else numpy.zeros((s,)))
        for sa, s in zip(selected_axes, image.shape)
    )
    for fc, x in zip(freq_cutoff, numpy.meshgrid(*axis_grid, indexing='ij')):
        f += (x / fc) ** 2
    return f


def _apw(freq_cutoff, max_padding):
    return min(max_padding, max(1, int(1.0 / (1e-10 + freq_cutoff))))


def _filter(image_f, f, order):
    image_f *= (1 + numpy.sqrt(f) ** (2 * order)) ** (-0.5)
    return image_f



import numpy
from scipy.ndimage import gaussian_filter
from skimage.data import binary_blobs
from skimage.util import random_noise

from aydin.it.transforms.fixedpattern import FixedPatternTransform


def add_patterned_noise(image, n):
    image = image.copy()
    image *= 1 + 0.1 * (numpy.random.rand(n, n) - 0.5)
    image += 0.1 * numpy.random.rand(n, n)
    # image += 0.1*numpy.random.rand(n)[]
    image = random_noise(image, mode="gaussian", var=0.00001, seed=0)
    image = random_noise(image, mode="s&p", amount=0.000001, seed=0)
    return image


def test_fixed_pattern_real():
    n = 128
    image = binary_blobs(length=n, seed=1, n_dim=3, volume_fraction=0.01).astype(
        numpy.float32
    )
    image = gaussian_filter(image, sigma=4)
    noisy = add_patterned_noise(image, n).astype(numpy.float32)

    bs = FixedPatternTransform(sigma=0)

    preprocessed = bs.preprocess(noisy)
    postprocessed = bs.postprocess(preprocessed)

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(noisy, name='noisy')
    #     viewer.add_image(preprocessed, name='preprocessed')
    #     viewer.add_image(postprocessed, name='postprocessed')

    assert image.shape == postprocessed.shape
    assert image.dtype == postprocessed.dtype
    assert numpy.abs(preprocessed - image).mean() < 0.007

    assert preprocessed.dtype == postprocessed.dtype
    assert numpy.abs(postprocessed - noisy).mean() < 1e-8

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(noisy, name='noisy')
    #     viewer.add_image(corrected, name='corrected')



from functools import partial
from typing import Optional

import numpy
import numpy as np
from skimage.restoration import denoise_wavelet as skimage_denoise_wavelet
from skimage.restoration import estimate_sigma

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.j_invariance.j_invariant_classic import calibrate_denoiser_classic


def calibrate_denoise_wavelet(
    image,
    crop_size_in_voxels: Optional[int] = None,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates a <a href="https://en.wikipedia.org/wiki/Wavelet_transform
    ">wavelet</a> denoiser for the given image and returns the optimal
    parameters obtained using the N2S loss.

    Note: we use the scikt-image implementation of wavelet denoising.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate wavelet denoiser for.

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate denoiser.
        (advanced)

    display_images: bool
        When True the denoised images encountered during optimisation are shown

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function and dictionary containing optimal parameters.

    """
    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # We make a first estimate of the noise sigma:
    estimated_sigma = estimate_sigma(image)

    # Sigma range:
    sigma_range = estimated_sigma * np.arange(0.25, 4, 0.01)
    # sigma_range = (0.0, 10.0)

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {
        'sigma': sigma_range,
        'wavelet': [
            'db1',
            'db2',
            'haar',
            'bior4.4',  # same as CDF 9/7 from JPEG 2000 lossy
            # 'sym9',
            # 'coif1',
            # 'coif5',
            # 'dmey',
        ],  # 'bior2.2', 'bior3.1', 'bior3.3',
        'mode': ['soft'],
        'method': ['BayesShrink', 'VisuShrink'],
    }

    # Partial function:
    _denoise_wavelet = partial(denoise_wavelet, **other_fixed_parameters)

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_classic(
            crop,
            _denoise_wavelet,
            denoise_parameters=parameter_ranges,
            display_images=display_images,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = image.nbytes * 3  # transform

    return denoise_wavelet, best_parameters, memory_needed


def denoise_wavelet(
    image,
    wavelet: str = 'db1',
    sigma: float = None,
    mode: str = 'soft',
    method: str = 'BayesShrink',
    **kwargs,
):
    """
    Denoises the given image using the scikit-image
    implementation of <a href="https://en.wikipedia.org/wiki/Wavelet_transform ">
    wavelet</a> denoising.
    \n\n
    Note: we use the scikt-image implementation of wavelet denoising.

    Parameters
    ----------
    image: ArrayLike
        Image to denoise

    wavelet : string, optional
        The type of wavelet to perform and can be any of the options
        ``pywt.wavelist`` outputs. The default is `'db1'`. For example,
        ``wavelet`` can be any of ``{'db2', 'haar', 'sym9'}`` and many more
          (see PyWavelets documentation).

    sigma : float or list, optional
        The noise standard deviation used when computing the wavelet detail
        coefficient threshold(s). When None (default), the noise standard
        deviation is estimated via the method in [2]_.

    mode : {'soft', 'hard'}, optional
        An optional argument to choose the type of denoising performed. It
        noted that choosing soft thresholding given additive noise finds the
        best approximation of the original image.

    method : {'BayesShrink', 'VisuShrink'}, optional
        Thresholding method to be used. The currently supported methods are
        "BayesShrink" [1]_ and "VisuShrink" [2]_. Defaults to "BayesShrink".

    kwargs : dict
        Any other parameters to be passed to scikit-image implementations

    Returns
    -------
    Denoised image as ndarray

    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    return skimage_denoise_wavelet(
        image, wavelet=wavelet, sigma=sigma, mode=mode, method=method, **kwargs
    )



import numpy
from skimage.data import camera

from aydin.util.misc.slicing_helper import apply_slicing


def test_apply_slicing():
    image = camera().astype(numpy.float32)

    assert image[10:20, 32:57].all() == apply_slicing(image, "[10:20, 32:57]").all()



from abc import ABC, abstractmethod
from typing import Sequence, Tuple


class FeatureGroupBase(ABC):
    """
    Feature Group base class
    """

    def __init__(self):
        """
        Constructs a feature group
        """

    @property
    @abstractmethod
    def receptive_field_radius(self) -> int:
        """
        Returns the receptive field radius in voxels

        Parameters
        ----------

        Returns
        -------
        result : int
            receptive field radius in pixels

        """
        raise NotImplementedError()

    @abstractmethod
    def num_features(self, ndim: int) -> int:
        """
        Returns the number of features given an image dimension

        Parameters
        ----------
        ndim : Number of dimension sof image for which features will be computed

        Returns
        -------
        result : int
            Number of features

        """
        raise NotImplementedError()

    @abstractmethod
    def prepare(
        self, image, excluded_voxels: Sequence[Tuple[int, ...]] = None, **kwargs
    ):
        """
        Prepares the computation of the features in the group.
        Sets the image for which features should be computed.

        Parameters
        ----------
        image : Image for which features will be computed
        excluded_voxels : voxels to exclude from feature as list of coordinate tuples.
        kwargs : key-value arguments for feature functions

        Returns
        -------

        """
        raise NotImplementedError()

    @abstractmethod
    def compute_feature(self, index: int, feature):
        """
        Computes feature of a given index. The feature index must be strictly less than the number of features returned
         by get_num_features.

        Parameters
        ----------
        index : index for feature
        feature : array into which to store feature

        Returns
        -------


        """
        raise NotImplementedError()

    def finish(self):
        """
        Cleans up and frees any resource allocated during feature computation. After cleanup this feature group can be reused to compute features for a new image.

        Parameters
        ----------

        Returns
        -------


        """
        # By default there is nothing to free:
        pass



from aydin.it.normalisers.base import NormaliserBase


class IdentityNormaliser(NormaliserBase):
    """Identity Normaliser"""

    def __init__(self, **kwargs):
        """Constructs a normalisers"""
        super().__init__(**kwargs)

    def calibrate(self, array):
        """Calibrate method

        Parameters
        ----------
        array : numpy.ndarray

        """
        self.original_dtype = array.dtype

        return None, None



# flake8: noqa
import numpy
from scipy.fft import dct
from scipy.signal import convolve2d
from skimage.exposure import rescale_intensity
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim
from skimage.restoration import richardson_lucy

from aydin.features.fast.fast_features import FastMultiscaleConvolutionalFeatures
from aydin.io.datasets import add_noise, newyork
from aydin.it.it_classic import ImageTranslatorClassic
from aydin.regression.gbm import GBMRegressor
from aydin.util.psf.simple_microscope_psf import SimpleMicroscopePSF


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def dctpsnr(true_image, test_image):
    true_image = n(dct(dct(true_image, axis=0), axis=1))
    test_image = n(dct(dct(test_image, axis=0), axis=1))

    return psnr(true_image, test_image)


# Prepare image:
image = newyork()
image = image.astype(numpy.float32)
image = n(image)

# Prepare PSF:
psf = SimpleMicroscopePSF()
psf_xyz_array = psf.generate_xyz_psf(dxy=0.406, dz=0.406, xy_size=17, z_size=17)
print(psf_xyz_array.shape)
kernel_psf = psf_xyz_array[0]

# degrade image with blurr and noise;
blurred_image = convolve2d(image, kernel_psf, 'same')
noisy_and_blurred_image = add_noise(
    blurred_image, intensity=10, variance=0.01, sap=0.00
)  # , intensity=100, variance=0.01, sap=0.001

# try classic_denoisers LR:
lr_deconvolved_image = richardson_lucy(
    noisy_and_blurred_image, kernel_psf, iterations=60
)

it = ImageTranslatorClassic(
    feature_generator=FastMultiscaleConvolutionalFeatures(exclude_scale_one=False),
    regressor=GBMRegressor(n_estimators=1000),
    normaliser_type='identity',
)

# lr deconvolution followed by denoising:
it.train(lr_deconvolved_image)
deconvolved_denoised_image = it.translate(lr_deconvolved_image)

# Try denoising and then lr deconvolution:
it.train(noisy_and_blurred_image)
denoised_image = it.translate(noisy_and_blurred_image)
denoised_deconvolved_image = richardson_lucy(denoised_image, kernel_psf, iterations=30)

# Train to translate input to denoised deconvolved, but without blind spot:
it = ImageTranslatorClassic(
    feature_generator=FastMultiscaleConvolutionalFeatures(exclude_scale_one=False),
    regressor=GBMRegressor(n_estimators=10),
    normaliser_type='identity',
)
it.train(noisy_and_blurred_image, denoised_deconvolved_image, force_jinv=False)
restored_image = it.translate(noisy_and_blurred_image)

# lr_reg_deconvolved_image = richardson_lucy_reg(noisy_and_blurred_image, kernel_psf, iterations=10)

# Clipping for comparison:
# lr_reg_deconvolved_image = numpy.clip(lr_reg_deconvolved_image, 0, 1)
lr_deconvolved_image = numpy.clip(lr_deconvolved_image, 0, 1)
denoised_image = numpy.clip(denoised_image, 0, 1)
deconvolved_denoised_image = numpy.clip(deconvolved_denoised_image, 0, 1)
denoised_deconvolved_image = numpy.clip(denoised_deconvolved_image, 0, 1)
restored_image = numpy.clip(restored_image, 0, 1)

# Compare results:
# print(
#     "lr_reg_deconvolved_image",
#     psnr(image, lr_reg_deconvolved_image),
#     dctpsnr(image, lr_reg_deconvolved_image),
#     ssim(image, lr_reg_deconvolved_image),
# )
print(
    "lr_deconvolved_image",
    psnr(image, lr_deconvolved_image),
    dctpsnr(image, lr_deconvolved_image),
    ssim(image, lr_deconvolved_image),
)
print(
    "denoised_image",
    psnr(image, denoised_image),
    dctpsnr(image, denoised_image),
    ssim(image, denoised_image),
)
print(
    "deconvolved_denoised_image",
    psnr(image, deconvolved_denoised_image),
    dctpsnr(image, deconvolved_denoised_image),
    ssim(image, deconvolved_denoised_image),
)
print(
    "denoised_deconvolved_image",
    psnr(image, denoised_deconvolved_image),
    dctpsnr(image, denoised_deconvolved_image),
    ssim(image, denoised_deconvolved_image),
)
print(
    "restored_image",
    psnr(image, restored_image),
    dctpsnr(image, restored_image),
    ssim(image, restored_image),
)

import napari

with napari.gui_qt():
    viewer = napari.Viewer()
    viewer.add_image(image, name='image')
    viewer.add_image(blurred_image, name='blurred_image')
    viewer.add_image(noisy_and_blurred_image, name='noisy_and_blurred_image')
    # viewer.add_image(lr_reg_deconvolved_image, name='lr_reg_deconvolved_image')
    viewer.add_image(lr_deconvolved_image, name='lr_deconvolved_image')
    viewer.add_image(denoised_image, name='denoised_image')
    viewer.add_image(deconvolved_denoised_image, name='deconvolved_denoised_image')
    viewer.add_image(denoised_deconvolved_image, name='denoised_deconvolved_image')
    viewer.add_image(restored_image, name='restored_image')



import numpy
import pytest
from scipy.ndimage import uniform_filter
from skimage.exposure import rescale_intensity

from aydin.util.fast_uniform_filter.numba_cpu_uf import numba_cpu_uniform_filter
from aydin.util.fast_uniform_filter.numba_gpu_uf import numba_gpu_uniform_filter
from aydin.util.fast_uniform_filter.parallel_uf import parallel_uniform_filter
from aydin.util.fast_uniform_filter.scipy_uf import scipy_uniform_filter
from aydin.io.datasets import newyork, examples_single


def test_uniform_filter_type_support():
    _test_compute_uniform_filter_type_support(scipy_uniform_filter)
    _test_compute_uniform_filter_type_support(parallel_uniform_filter)
    _test_compute_uniform_filter_type_support(numba_cpu_uniform_filter)


@pytest.mark.gpu
def test_uniform_filter_type_support_gpu():
    _test_compute_uniform_filter_type_support(numba_gpu_uniform_filter)


def test_uniform_filter_different_sizes():
    _test_compute_uniform_filter_different_sizes(scipy_uniform_filter)
    _test_compute_uniform_filter_different_sizes(parallel_uniform_filter)
    _test_compute_uniform_filter_different_sizes(numba_cpu_uniform_filter)


@pytest.mark.gpu
def test_uniform_filter_different_sizes_gpu():
    _test_compute_uniform_filter_different_sizes(numba_gpu_uniform_filter)


def test_compute_uniform_filter_1d():
    _test_compute_uniform_filter_1d(scipy_uniform_filter)
    _test_compute_uniform_filter_1d(parallel_uniform_filter)
    _test_compute_uniform_filter_1d(numba_cpu_uniform_filter)


@pytest.mark.gpu
def test_compute_uniform_filter_1d_gpu():
    _test_compute_uniform_filter_1d(numba_gpu_uniform_filter)


def test_compute_uniform_filter_2d():
    _test_compute_uniform_filter_2d(scipy_uniform_filter)
    _test_compute_uniform_filter_2d(parallel_uniform_filter)
    _test_compute_uniform_filter_2d(numba_cpu_uniform_filter)


@pytest.mark.gpu
def test_compute_uniform_filter_2d_gpu():
    _test_compute_uniform_filter_2d(numba_gpu_uniform_filter)


def test_compute_uniform_filter_3d():
    _test_compute_uniform_filter_3d(scipy_uniform_filter)
    _test_compute_uniform_filter_3d(parallel_uniform_filter)
    _test_compute_uniform_filter_3d(numba_cpu_uniform_filter)


@pytest.mark.gpu
def test_compute_uniform_filter_3d_gpu():
    _test_compute_uniform_filter_3d(numba_gpu_uniform_filter)


def _normalise(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def _test_compute_uniform_filter_type_support(_fun_):
    image = newyork()

    _run_test_for_type(_fun_, image.astype(dtype=numpy.float32))
    _run_test_for_type(_fun_, image.astype(dtype=numpy.float16), decimal=1)
    _run_test_for_type(_fun_, image.astype(dtype=numpy.uint32), decimal=0)
    _run_test_for_type(_fun_, image.astype(dtype=numpy.uint16), decimal=0)
    _run_test_for_type(_fun_, image.astype(dtype=numpy.uint8), decimal=0)


def _test_compute_uniform_filter_different_sizes(_fun_):
    sizes = [1, 2, 3, 4, 5, 7, 8, 16, 32, 128]

    for size in sizes:
        _test_compute_uniform_filter_2d(_fun_, size=size)

    sizes = [(1, 2), (16, 7), (3, 7), (4, 4)]

    for size in sizes:
        _test_compute_uniform_filter_2d(_fun_, size=size)


def _run_test_for_type(_fun_, image, decimal=3, size=3):
    filtered_image = _fun_(image, size=size)
    assert filtered_image.dtype == image.dtype

    filtered_image = filtered_image.astype(dtype=numpy.float32, copy=False)
    scipy_filtered_image = uniform_filter(
        image.astype(dtype=numpy.float32, copy=False), size=size, mode="nearest"
    ).astype(dtype=numpy.float32, copy=False)
    numpy.testing.assert_array_almost_equal(
        filtered_image, scipy_filtered_image, decimal=decimal
    )


def _test_compute_uniform_filter_1d(_fun_, size=3):
    image = _normalise(newyork().astype(numpy.float32))
    image = image[512, :]

    filtered_image = _fun_(image, size=size)
    scipy_filtered_image = uniform_filter(image, size=size, mode="nearest")

    numpy.testing.assert_array_almost_equal(
        filtered_image, scipy_filtered_image, decimal=3
    )


def _test_compute_uniform_filter_2d(_fun_, size=3):
    image = _normalise(newyork().astype(numpy.float32))

    filtered_image = _fun_(image, size=size)
    scipy_filtered_image = uniform_filter(image, size=size, mode="nearest")

    numpy.testing.assert_array_almost_equal(
        filtered_image, scipy_filtered_image, decimal=3
    )


def _test_compute_uniform_filter_3d(_fun_, size=3):
    islet = examples_single.royerlab_hcr.get_array().squeeze()
    image = islet[2, :60, 0 : 0 + 1524, 0 : 0 + 1524]

    filtered_image = _fun_(image, size=size)
    scipy_filtered_image = uniform_filter(image, size=size, mode="nearest")

    numpy.testing.assert_array_almost_equal(
        filtered_image, scipy_filtered_image, decimal=3
    )


def _test_compute_uniform_filter_4d(_fun_, size=3):
    image = examples_single.hyman_hela.get_array().squeeze()
    image = image[..., 0:128, 0:128]

    filtered_image = _fun_(image, size=size)
    scipy_filtered_image = uniform_filter(image, size=size, mode="nearest")

    numpy.testing.assert_array_almost_equal(
        filtered_image, scipy_filtered_image, decimal=3
    )



import numpy

from aydin.analysis.camera_simulation import simulate_camera_image
from aydin.io.datasets import characters
from aydin.it.transforms.variance_stabilisation import VarianceStabilisationTransform


def test_vst():
    image = characters()
    image = image.astype(numpy.float32) * 0.1
    noisy = simulate_camera_image(image)

    for mode in [
        'box-cox',
        'yeo-johnson',
        'quantile',
        'anscomb',
        'log',
        'sqrt',
        'identity',
    ]:
        print(f"testing mode: {mode}")
        vst = VarianceStabilisationTransform(mode=mode, leave_as_float=False)

        preprocessed = vst.preprocess(noisy)
        postprocessed = vst.postprocess(preprocessed)

        # import napari
        # with napari.gui_qt():
        #     viewer = napari.Viewer()
        #     viewer.add_image(image, name='image')
        #     viewer.add_image(noisy, name='noisy')
        #     viewer.add_image(preprocessed, name='preprocessed')
        #     viewer.add_image(postprocessed, name='postprocessed')

        error = numpy.abs(
            postprocessed.astype(numpy.float32) - noisy.astype(numpy.float32)
        ).mean()

        print(f"round-trip error: {error}")

        assert error < 0.33

        assert postprocessed.dtype == noisy.dtype
        assert postprocessed.shape == noisy.shape



import numpy
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.analysis.image_metrics import mutual_information, spectral_mutual_information
from aydin.io.datasets import characters, add_blur_2d, add_noise, normalise
from aydin.restoration.deconvolve.lr import LucyRichardson


def printscore(header, val1, val2, val3, val4):
    print(f"{header}: \t {val1:.4f} \t {val2:.4f} \t {val3:.4f} \t {val4:.4f}")


def test_run():
    # Prepare the input image
    image = normalise(characters().astype(numpy.float32))
    blurred_image, psf_kernel = add_blur_2d(image)
    noisy_and_blurred_image = add_noise(
        blurred_image, intensity=10000, variance=0.0001, sap=0.0000001
    )

    # Call the LucyRichardson restoration
    lr = LucyRichardson(
        psf_kernel=psf_kernel, max_num_iterations=20, backend='scipy-cupy'
    )

    lr.train(noisy_and_blurred_image, noisy_and_blurred_image)

    lr_deconvolved_image = lr.deconvolve(noisy_and_blurred_image)

    print("Below in order: PSNR, norm spectral mutual info, norm mutual info, SSIM: ")
    printscore(
        "n&b image",
        psnr(image, noisy_and_blurred_image),
        spectral_mutual_information(image, noisy_and_blurred_image),
        mutual_information(image, noisy_and_blurred_image),
        ssim(image, noisy_and_blurred_image),
    )

    printscore(
        "lr      ",
        psnr(image, lr_deconvolved_image),
        spectral_mutual_information(image, lr_deconvolved_image),
        mutual_information(image, lr_deconvolved_image),
        ssim(image, lr_deconvolved_image),
    )

    assert psnr(image, noisy_and_blurred_image) < psnr(image, lr_deconvolved_image)
    assert mutual_information(image, noisy_and_blurred_image) < mutual_information(
        image, lr_deconvolved_image
    )
    assert ssim(image, noisy_and_blurred_image) < ssim(image, lr_deconvolved_image)
    assert 19.0 < psnr(image, lr_deconvolved_image)



import numpy
from scipy.ndimage import convolve
from skimage.exposure import rescale_intensity

from aydin.features.groups.convolutional import ConvolutionalFeatures
from aydin.io.datasets import camera


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def test_convolutional_feature_group():
    # get image:
    image = n(camera().astype(numpy.float32))

    # Instantiates convolutional features:
    ones = numpy.ones(shape=(3, 3))
    twos = 2 * numpy.ones(shape=(3, 3))
    convolutions = ConvolutionalFeatures(kernels=[ones, twos])
    assert convolutions.num_features(image.ndim) == 2

    # Check receptive field radius:
    assert convolutions.receptive_field_radius == 1

    # Set image:
    convolutions.prepare(image, [])

    # compute features and check their valididty:
    feature = numpy.empty_like(image)

    # Compute first convolution:
    convolutions.compute_feature(index=0, feature=feature)
    assert (feature == convolve(image, weights=ones)).all()

    # Compute second convolution:
    convolutions.compute_feature(index=1, feature=feature)
    assert (feature == convolve(image, weights=twos)).all()



import os
from abc import ABC, abstractmethod
from os.path import join
import jsonpickle
import numpy

from aydin.util.misc.json import encode_indent
from aydin.util.log.log import lprint, lsection
from aydin.util.offcore.offcore import offcore_array


class FeatureGeneratorBase(ABC):
    """
    Feature Generator base class
    """

    _max_non_batch_dims = 4
    _max_voxels = 512 ** 3

    def __init__(self):
        """
        Constructs a feature generator
        """

        self.check_nans = False
        self.debug_force_memmap = False

        # Implementations must initialise the dtype so that feature arrays can be created with correct type:
        self.dtype = None

    def save(self, path: str):
        """
        Saves a 'all-batteries-inlcuded' feature generator at a given path (folder)

        Parameters
        ----------
        path : str
            path to save to

        Returns
        -------
        frozen

        """

        os.makedirs(path, exist_ok=True)

        frozen = encode_indent(self)

        lprint(f"Saving feature generator to: {path}")
        with open(join(path, "feature_generation.json"), "w") as json_file:
            json_file.write(frozen)

        return frozen

    @staticmethod
    def load(path: str):
        """
        Returns a 'all-batteries-inlcuded' feature generator from a given path (folder)

        Parameters
        ----------
        path : str
            path to load from

        Returns
        -------
        thawed

        """

        lprint(f"Loading feature generator from: {path}")
        with open(join(path, "feature_generation.json"), "r") as json_file:
            frozen = json_file.read()

        thawed = jsonpickle.decode(frozen)

        thawed._load_internals(path)

        return thawed

    @abstractmethod
    def _load_internals(self, path: str):
        raise NotImplementedError()

    @abstractmethod
    def get_receptive_field_radius(self):
        """
        Returns the receptive field radius in pixels
        """
        raise NotImplementedError()

    @abstractmethod
    def compute(
        self,
        image,
        exclude_center_feature=False,
        exclude_center_value=False,
        features=None,
        feature_last_dim=True,
        passthrough_channels=None,
        num_reserved_features=0,
        excluded_voxels=None,
        spatial_feature_offset=None,
        spatial_feature_scale=None,
    ):
        """
        Computes the features given an image. If the input image is of shape (d,h,w),
        resulting features are of shape (n,d,h,w) where n is the number of features.

        Parameters
        ----------
        image : numpy.ndarray
            image for which features are computed
        exclude_center_feature : bool
        exclude_center_value : bool
        features
        feature_last_dim : bool
        passthrough_channels
        num_reserved_features : int
        excluded_voxels
        spatial_feature_offset
        spatial_feature_scale

        Returns
        -------
        feature array : numpy.ndarray

        """
        raise NotImplementedError()

    def create_feature_array(self, image, nb_features):
        """
        Creates a feature array of the right size and possibly in a 'lazy' way using memory mapping.

        Parameters
        ----------
        image : numpy.ndarray
            image for which features are created
        nb_features : int

        Returns
        -------
        feature array : numpy.ndarray

        """

        with lsection(f'Creating feature array for image of shape: {image.shape}'):
            # That's the shape we need:
            shape = (nb_features, image.shape[0]) + image.shape[2:]
            dtype = image.dtype if self.dtype is None else self.dtype
            dtype = numpy.float32 if dtype == numpy.float16 else dtype
            array = offcore_array(
                shape=shape, dtype=dtype, force_memmap=self.debug_force_memmap
            )
            return array



import time
from os.path import join

import numpy
from skimage.exposure import rescale_intensity

from aydin.io import imread
from aydin.io.datasets import examples_single
from aydin.io.folders import get_temp_folder
from aydin.it.normalisers.base import NormaliserBase
from aydin.it.normalisers.identity import IdentityNormaliser
from aydin.it.normalisers.minmax import MinMaxNormaliser
from aydin.it.normalisers.percentile import PercentileNormaliser


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def test_percentile_saveload():
    saveload(PercentileNormaliser(0))


def test_minmax_saveload():
    saveload(MinMaxNormaliser())


def test_identity_saveload():
    saveload(IdentityNormaliser())


def saveload(normaliser):
    input_path = examples_single.hyman_hela.get_path()
    array, metadata = imread(input_path)
    assert array.dtype == numpy.uint16

    normaliser.calibrate(array)
    print(f"Before normalisation: min,max = {(array.min(), array.max())}")

    temp_file = join(
        get_temp_folder(), "test_normaliser_saveload.json" + str(time.time())
    )
    normaliser.save(temp_file)
    del normaliser

    loaded_normaliser = NormaliserBase.load(temp_file)

    new_array = array.copy()
    normalised_array = loaded_normaliser.normalise(new_array)
    print(
        f"After normalisation: min,max = {(normalised_array.min(), normalised_array.max())}"
    )
    assert normalised_array.dtype == numpy.float32

    denormalised_array = loaded_normaliser.denormalise(normalised_array)
    print(
        f"After denormalisation: min,max = {(denormalised_array.min(), denormalised_array.max())}"
    )
    assert denormalised_array.dtype == numpy.uint16

    assert (
        abs(array.min() - denormalised_array.min()) < 5
        and abs(array.max() - denormalised_array.max()) < 20
    )



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QCheckBox,
    QLabel,
    QVBoxLayout,
    QScrollArea,
)

from aydin.gui._qt.custom_widgets.constructor_arguments import (
    ConstructorArgumentsWidget,
)
from aydin.gui._qt.job_runners.preview_job_runner import PreviewJobRunner
from aydin.gui._qt.custom_widgets.horizontal_line_break_widget import (
    QHorizontalLineBreakWidget,
)
from aydin.gui._qt.custom_widgets.vertical_line_break_widget import (
    QVerticalLineBreakWidget,
)
from aydin.util.string.break_text import break_text


class TransformsTabItem(QWidget):
    def __init__(
        self,
        parent,
        name=None,
        arg_names=None,
        arg_defaults=None,
        arg_annotations=None,
        transform_class=None,
    ):
        super(TransformsTabItem, self).__init__(parent)

        self.parent = parent
        self.arg_names = arg_names
        self.annotations = []
        self.transform_class = transform_class
        self.line_edits = []

        self.layout = QHBoxLayout()

        explanation_text_string = self.transform_class.__doc__
        explanation_text_string = break_text(explanation_text_string)
        explanation_text_string = explanation_text_string.replace('\n', '<br>')
        self.explanation_text = QLabel(explanation_text_string, self)
        self.explanation_text.setTextFormat(Qt.RichText)
        self.explanation_text.setOpenExternalLinks(True)
        # self.explanation_text.setWordWrap(True)
        self.explanation_text.setAlignment(Qt.AlignTop)
        self.layout.addWidget(self.explanation_text, 45)

        # Vertical Line Break
        self.layout.addWidget(QVerticalLineBreakWidget(self))

        self.transform_details_layout = QVBoxLayout()
        self.transform_details_layout.setAlignment(Qt.AlignTop)

        # Checkbox
        self.enabling_checkboxes_layout = QHBoxLayout()

        self.preprocess_checkbox = QCheckBox("Enable" if name is None else name)
        self.enabling_checkboxes_layout.addWidget(self.preprocess_checkbox)
        self.enabling_checkboxes_layout.addStretch()

        self.preview_job_runner = PreviewJobRunner(
            self, self.parent.parent.parent.threadpool
        )
        self.enabling_checkboxes_layout.addWidget(self.preview_job_runner)

        self.transform_details_layout.addLayout(self.enabling_checkboxes_layout)

        # Postprocess checkbox
        self.postprocess_checkbox = QCheckBox(
            self.transform_class.postprocess_description
        )
        self.postprocess_checkbox.setEnabled(
            self.transform_class.postprocess_supported
            and self.preprocess_checkbox.isChecked()
        )
        self.postprocess_checkbox.setChecked(
            self.transform_class.postprocess_recommended
        )
        self.preprocess_checkbox.stateChanged.connect(
            self.preprocess_chechbox_on_state_changed
        )
        self.preprocess_chechbox_on_state_changed()
        self.transform_details_layout.addWidget(self.postprocess_checkbox)

        self.transform_details_layout.addWidget(QHorizontalLineBreakWidget(self))

        # Parameters
        self.scroll = QScrollArea()
        self.scroll.setStyleSheet("QScrollArea {border: none;}")
        self.scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.constructor_arguments_widget = ConstructorArgumentsWidget(
            self,
            arg_names=arg_names,
            arg_defaults=arg_defaults,
            arg_annotations=arg_annotations,
            reference_class=transform_class,
        )
        self.scroll.setWidget(self.constructor_arguments_widget)
        self.scroll.setWidgetResizable(True)
        self.scroll.setMinimumHeight(200)
        self.transform_details_layout.addWidget(self.scroll)

        self.layout.addLayout(self.transform_details_layout, 45)
        self.layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.layout)

    def preprocess_chechbox_on_state_changed(self):
        self.postprocess_checkbox.setEnabled(self.preprocess_checkbox.isChecked())

    @property
    def params_dict(self):
        if self.preprocess_checkbox.isChecked():
            params_dict = self.constructor_arguments_widget.params_dict
            params_dict["kwargs"][
                "do_postprocess"
            ] = self.postprocess_checkbox.isChecked()

            return params_dict



# flake8: noqa
import time

import numpy
from skimage.metrics import structural_similarity as ssim

from aydin.io.datasets import normalise, add_noise, characters, add_blur_2d
from aydin.it.deconvolution.lr_deconv_cupy import ImageTranslatorLRDeconvCupy
from aydin.util.log.log import lsection, lprint, Log


def test():
    Log.enable_output = True

    image = characters()
    image = normalise(image.astype(numpy.float32))
    blurred_image, psf_kernel = add_blur_2d(image)
    noisy_and_blurred_image = add_noise(
        blurred_image, intensity=10000, variance=0.0001, sap=0.0000001
    )

    lr = ImageTranslatorLRDeconvCupy(psf_kernel=psf_kernel, max_num_iterations=20)

    with lsection("training:"):
        start = time.time()
        lr.train(noisy_and_blurred_image, noisy_and_blurred_image)
        stop = time.time()
        elapsed_train = stop - start

    with lsection("translating:"):
        start = time.time()
        lr_deconvolved_image = lr.translate(noisy_and_blurred_image)
        stop = time.time()
        elapsed_translate = stop - start

    lprint(f"elapsed_train={elapsed_train}")
    lprint(f"elapsed_translate={elapsed_translate}")

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(blurred_image, name='blurred')
    #     viewer.add_image(noisy_and_blurred_image, name='noisy')
    #     viewer.add_image(lr_deconvolved_image, name='lr_deconvolved_image')

    assert elapsed_train < 0.1
    assert elapsed_translate < 0.7

    assert ssim(image, noisy_and_blurred_image) < 0.75
    assert ssim(image, lr_deconvolved_image) > 0.89



import math
from functools import partial
from typing import Callable, Dict, Union, Tuple, List, Any
import numpy
from skimage.metrics import mean_squared_error

from aydin.util.j_invariance.j_invariant_classic import (
    _j_invariant_loss,
    _product_from_dict,
    _generate_mask,
)
from aydin.util.log.log import lsection, lprint
from aydin.util.optimizer.optimizer import Optimizer


def calibrate_denoiser_smart(
    image,
    denoise_function: Callable,
    denoise_parameters: Dict[str, Union[Tuple[float, float], List[Any]]],
    max_num_evaluations: int = 128,
    patience: int = 64,
    stride: int = 4,
    loss_function: Callable = mean_squared_error,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates denoiser using self-supervised loss from Batson & Royer*
    Derived from code here:
    https://scikit-image.org/docs/dev/auto_examples/filters/plot_j_invariant_tutorial.html
    Reference: "Noise2Self: Blind Denoising by Self-Supervision, International
    Conference on Machine Learning, p. 524-533 (2019)"

    This 'smart' version uses a fancy optimizer to try to minimise the number
    of evaluations needed.

    Parameters
    ----------
    image: ArrayLike
        Image to calibate denoiser with.
    denoise_function: Callable
        Denosing function to calibrate. Should take an image as first parameter,
        all other parameters should have defaults
    denoise_parameters:
        Dictionary with keys corresponding to parameters of the denoising function.
        Values are either: (i) a list of possible values (categorical parameter),
        or (ii) a tuple of floats defining the bounds of that numerical parameter.
    max_num_evaluations: int
        Max number of function evaluations. This is per the size of the cartesian
        product of categorical parameters.
    patience : int
        After 'patience' evaluations we stop the optimiser
    stride: int
        Stride to compute self-supervised loss.
    loss_function: Callable
        Loss/Error function: takes two arrays and returns a distance-like function.
        Can be:  structural_error, mean_squared_error, _mean_absolute_error
    display_images: bool
        If True the denoised images for each parameter tested are displayed.
        this _will_ be slow.
    other_fixed_parameters: dict
        Other fixed parameters to pass to the denoiser function.


    Returns
    -------
    Dictionary with optimal parameters

    """

    # Pass fixed parameters:
    denoise_function = partial(denoise_function, **other_fixed_parameters)

    # for display purposes:
    denoised_images = []

    with lsection("Calibrating denoiser:"):

        # Generate mask:
        mask = _generate_mask(image, stride)

        # first we separate the categorical from numerical parameters;
        categorical_parameters = {}
        numerical_parameters = {}
        for name, value in denoise_parameters.items():
            if type(value) is list:
                # This is a categorical parameter:
                categorical_parameters[name] = value
            elif type(value) is tuple:
                # This is a numerical parameter:
                numerical_parameters[name] = value

        # Let's fix the order of numerical parameters:
        numerical_parameters_names = list(numerical_parameters.keys())
        numerical_parameters_bounds = list(
            numerical_parameters[n] for n in numerical_parameters_names
        )
        lprint(f"Parameter names : {numerical_parameters_names}")
        lprint(f"Parameter bounds: {numerical_parameters_bounds}")

        # number of numerical parameters:
        n = len(numerical_parameters_names)

        # Lets' expand the categorical parameters into their cartesian product:
        expanded_categorical_parameters = list(
            _product_from_dict(categorical_parameters)
        )

        # We remember the best combination, parameters and values:
        best_combination = None
        best_numerical_parameters = None
        best_loss_value = -math.inf

        # we optimise for each such combination of categorical parameters:
        with lsection(f"Going through {n} combinations of categorical parameters:"):

            for combination in expanded_categorical_parameters:
                with lsection(
                    f"Optimising categorical parameters combination: {combination}"
                ):

                    # This is the fucntion to optimize:
                    def opt_function(*point):

                        # Given a point, we build the parameter dict:
                        param = dict(combination)
                        for name, value in zip(numerical_parameters_names, point):
                            param[name] = value

                        # We compute the J-inv loss:
                        loss = _j_invariant_loss(
                            image,
                            denoise_function,
                            mask=mask,
                            loss_function=loss_function,
                            denoiser_kwargs=param,
                        )

                        if math.isnan(loss) or math.isinf(loss):
                            loss = math.inf

                        # lprint(f"({point}) -> {loss}")
                        # denoise image and store it, if needed:
                        if display_images:
                            denoised_image = denoise_function(image, **param)
                            denoised_images.append(denoised_image)

                        # our optimizer is a maximiser!
                        return -loss

                    # We optimise here:
                    new_parameters, new_loss_value = Optimizer().optimize(
                        opt_function,
                        bounds=numerical_parameters_bounds,
                        max_num_evaluations=max_num_evaluations,
                        patience=patience,
                    )

                    # We check if this is, or not, the best combination to date:
                    if (
                        new_loss_value > best_loss_value
                    ):  # our optimizer is a maximiser!
                        best_combination = combination
                        best_loss_value = new_loss_value
                        best_numerical_parameters = new_parameters

                    lprint(
                        f"Best numerical parameters: {best_numerical_parameters} for combination: {combination}"
                    )

            # Let convert back the parameters into a dictionary:
            best_parameters = best_combination | {
                name: float(value)
                for name, value in zip(
                    numerical_parameters_names, best_numerical_parameters
                )
            }

            lprint(f"Best parameters: {best_parameters}")

    # Display if needed:
    if display_images:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='image')
            viewer.add_image(numpy.stack(denoised_images), name='denoised')

    return best_parameters | other_fixed_parameters



import pytest

from aydin.io.datasets import cropped_newyork
from aydin.it.classic_denoisers.demo.demo_2D_dictionary_fixed import (
    demo_dictionary_fixed,
)
from aydin.it.classic_denoisers.demo.demo_2D_dictionary_learned import (
    demo_dictionary_learned,
)
from aydin.it.classic_denoisers.dictionary_fixed import denoise_dictionary_fixed

from aydin.it.classic_denoisers.test.util_test_nd import check_nd


def test_dictionary_learned():
    assert demo_dictionary_learned(cropped_newyork(), display=False) >= 0.636 - 0.01


@pytest.mark.heavy
def test_dictionary_fixed():
    assert demo_dictionary_fixed(cropped_newyork(), display=False) >= 0.636 - 0.01


def test_dictionary_nd():
    check_nd(denoise_dictionary_fixed)



import numpy
import torch
import torch.nn as nn
import torch.nn.functional as F

"""
    JIDCnet2D -- J-invariant Dilated Convolution Net
"""


class JInet2D(nn.Module):
    def __init__(
        self,
        num_input_channels=1,
        num_output_channels=1,
        kernel_sizes=None,
        num_features=None,
        num_dense_layers=3,
        num_channels=None,
        final_relu=False,
        full_convolutional_across_channels=True,
        padding_mode='zeros',
        degressive_residuals=True,
        kernel_continuity_regularisation=True,
        sine_activations=True,
    ):
        """

        Parameters
        ----------
        num_input_channels
        num_output_channels
        kernel_sizes
        num_features
        num_dense_layers
        num_channels
        final_relu
        """
        super().__init__()

        # These are the scales and associated kernel sizes and number of features

        if kernel_sizes is None:
            # kernel_sizes = [5, 5, 5, 3, 3, 3, 3, 3]
            kernel_sizes = [7, 5, 3, 3, 3, 3, 3, 3]
        if num_features is None:
            # num_features = [24, 24, 24, 12, 12, 12, 2, 1]
            num_features = [64, 32, 16, 8, 4, 2, 1, 1]
        self.kernel_sizes = kernel_sizes
        self.num_features = num_features
        assert len(kernel_sizes) == len(num_features)

        # How many scales?
        self.num_scales = len(kernel_sizes)

        # final relu?
        self.final_relu = final_relu

        # Degressive residuals ?
        self.degressive_residuals = degressive_residuals

        # Enforce kernel continuity?
        self.kernel_continuity_regularisation = kernel_continuity_regularisation

        # Sine activations?
        self.sine_activations = sine_activations

        self.dilated_conv_list = nn.ModuleList()
        total_num_features = 0
        current_receptive_field_radius = 0
        current_num_channels = num_input_channels
        for scale_index in range(self.num_scales):
            # Get kernel size and number of features:
            size = kernel_sizes[scale_index]
            num = num_features[scale_index]

            # radius and dilation:
            radius = (size - 1) // 2
            dilation = 1 + current_receptive_field_radius

            # Do we go full convolutional across channels or do we try to minimise parameters?
            num_groups = (
                1
                if full_convolutional_across_channels
                else min(current_num_channels, num)
            )

            # Setup dilated convolution:
            dilated_convolution = nn.Conv2d(
                current_num_channels,
                num,
                kernel_size=size,
                padding=dilation * radius,
                dilation=dilation,
                padding_mode=padding_mode,
                groups=num_groups,
            )
            print(dilated_convolution)

            # We keep track of the total number of features until now:
            total_num_features += num

            # we update the current receptive field radius:
            current_receptive_field_radius += dilation * radius

            # next convolution num of input channels is this last convolution number of output channels:
            current_num_channels = num

            # append convolution ton the list:
            self.dilated_conv_list.append(dilated_convolution)

        # we save the receptive field radius:
        self.receptive_field_radius = current_receptive_field_radius

        # We keep the number of features:
        self.total_num_features = total_num_features

        # By default the number of channels for the deep layers is the total number of spatial features generated
        if num_channels is None:
            num_channels = 2 * total_num_features

        # Instanciates dense layers:
        self.dense_layers_list = nn.ModuleList()
        for level in range(0, num_dense_layers):
            # number of input and output channels:
            num_in = total_num_features if level == 0 else num_channels
            num_out = (
                num_output_channels if level == num_dense_layers - 1 else num_channels
            )

            # Setup of channel-wise dense layer (1x1 convolution)
            dense = nn.Conv2d(num_in, num_out, kernel_size=1, padding=0)
            self.dense_layers_list.append(dense)

        self.finalconv = nn.Conv2d(
            num_channels, num_output_channels, kernel_size=1, padding=0
        )

    def forward(self, x0):
        """

        Parameters
        ----------
        x0  input

        Returns output of model
        -------

        """
        # Deep spatial feature generation:
        x = x0
        features = []
        for scale_index in range(self.num_scales):
            dilated_convolution = self.dilated_conv_list[scale_index]
            x = dilated_convolution(x)
            features.append(x)
            x = F.leaky_relu(x, negative_slope=0.01)

        # stack all features into one tensor:
        x = torch.cat(features, 1)

        # Residual 'feed-forward' perceptron for inference, i.e. 1x1 convolutions
        y = None
        f = 1
        for dense in self.dense_layers_list:
            x = dense(x)
            if self.sine_activations:
                x = torch.sin(x)
            else:
                x = F.leaky_relu(x, negative_slope=0.01)
            if y is None:
                y = x
            else:
                y = y + f * x

            if self.degressive_residuals:
                # deeper layers are expected to contribute less and less to solution:
                f *= 0.5

        # final dense layer to bring down number of channels:
        y = self.finalconv(y)

        # final relu:
        if self.final_relu:
            y = torch.clamp(y, 0, 1)

        return y

    def enforce_blind_spot(self):
        """
        This must be called after each optimisation step to guarantee J-invariance
        """

        with torch.no_grad():
            for skipconv in self.dilated_conv_list[0:1]:  #
                indexes = tuple((i - 1) // 2 for i in skipconv.kernel_size)
                skipconv._parameters['weight'][:, :, indexes[0], indexes[1]] = 0

    def post_optimisation(self):
        """
        This must be called after each optimisation for kernel regularisation
        """

        b = 0.0005
        with torch.no_grad():

            if self.kernel_continuity_regularisation:
                for skipconv in self.dilated_conv_list:  # [0:2]
                    weights = skipconv._parameters['weight']
                    num_channels = weights.shape[1]
                    kernel = numpy.array([[b, b, b], [b, 1, b], [b, b, b]])
                    kernel = kernel[numpy.newaxis, numpy.newaxis, ...].astype(
                        numpy.float32
                    )
                    kernel = torch.from_numpy(kernel).to(weights.device)
                    kernel /= kernel.sum()
                    kernel = kernel.expand(num_channels, 1, -1, -1)
                    weights = F.conv2d(weights, kernel, groups=num_channels, padding=1)
                    skipconv._parameters['weight'] = weights

                    # decrease for upper layers:
                    b *= 0.5

    def fill_blind_spot(self):
        """
        This must be called after training to remove the blind-spot.
        """

        for skipconv in self.dilated_conv_list[0:1]:
            weights = skipconv._parameters['weight']
            num_channels = weights.shape[1]
            original_sum = weights.sum()
            b = 1
            kernel = numpy.array([[b, b, b], [b, 0, b], [b, b, b]])
            kernel = kernel[numpy.newaxis, numpy.newaxis, ...].astype(numpy.float32)
            kernel = torch.from_numpy(kernel).to(weights.device)
            kernel /= kernel.sum()
            kernel = kernel.expand(num_channels, 1, -1, -1)
            filtered_weights = F.conv2d(weights, kernel, groups=num_channels, padding=1)

            indexes = tuple((i - 1) // 2 for i in skipconv.kernel_size)
            weights[:, :, indexes[0], indexes[1]] = filtered_weights[
                :, :, indexes[0], indexes[1]
            ]
            weights *= original_sum / weights.sum()

            skipconv._parameters['weight'] = weights

    def visualise_weights(self):
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            for i, dilated_conv in enumerate(self.dilated_conv_list):  #
                array = dilated_conv._parameters['weight'].cpu().detach().numpy()
                viewer.add_image(array, name=f'layer{i}', rgb=False)



# flake8: noqa
import numpy
from scipy.ndimage import uniform_filter

from aydin.util.fast_uniform_filter import parallel_uniform_filter
from aydin.io.datasets import fibsem, examples_single
from aydin.util.log.log import lsection, Log


def demo_par_uniform(image_name: str, image, repeats=32):
    Log.enable_output = True

    size = 4

    with lsection(f"Par {image_name} (r={repeats}):"):
        for _ in range(repeats):
            par_filtered_image = parallel_uniform_filter(image, size=size)

    with lsection(f"Scipy {image_name} (r={repeats}):"):
        for _ in range(repeats):
            scipy_filtered_image = uniform_filter(image, size=size, mode="nearest")

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(
            par_filtered_image, name='par_filtered_image', colormap='plasma'
        )
        viewer.add_image(
            scipy_filtered_image, name='scipy_filtered_image', colormap='plasma'
        )
        viewer.add_image(
            numpy.abs(scipy_filtered_image - par_filtered_image),
            name='scipy_filtered_image',
            colormap='plasma',
        )

    numpy.testing.assert_array_almost_equal(
        par_filtered_image, scipy_filtered_image, decimal=3
    )


image = fibsem(full=True).astype(numpy.float32)
demo_par_uniform("fibsem 2D", image)

image = examples_single.royerlab_hcr.get_array().squeeze()  # [0, ..., :320]
demo_par_uniform("islet 3D", image, repeats=10)

image = examples_single.hyman_hela.get_array().squeeze()
demo_par_uniform("hela 4D", image, repeats=1)



import numpy
from skimage.exposure import rescale_intensity

from aydin.io import io
from aydin.io.datasets import examples_single
from aydin.it.cnn.util.data_util import random_sample_patches
import napari


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def demo():
    # Load image
    image_path = examples_single.royerlab_hcr.get_path()
    image0, metadata = io.imread(image_path)
    print(image0.shape)
    image0 = n(image0.squeeze())

    image0 = numpy.expand_dims(image0[1:2], -1)
    tile_size = (64, 64, 64)
    num_tile = 500
    adoption_rate = 0.5
    input_data = random_sample_patches(image0, tile_size, num_tile, adoption_rate)

    imgpatch_int = numpy.zeros(image0.shape)
    for i in input_data:
        imgpatch_int[i[:-1]] += 1
    image0 = rescale_intensity(image0.squeeze().astype(numpy.float64), out_range=(0, 1))
    imgpatch_int = rescale_intensity(imgpatch_int.squeeze(), out_range=(0, 1))
    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image0)
        viewer.add_image(imgpatch_int)


demo()



from functools import reduce
from math import ceil
import numpy


class BlindSpotTransform:
    """
    'BlindSpot' Transform (and inverse)

    Parameters
    ----------
    blind_spots : list

    """

    def __init__(self, blind_spots):
        super().__init__()

        self.blind_spots = blind_spots

        self.min_bounds = reduce(
            (lambda tx, ty: tuple(min(x, y) for x, y in zip(tx, ty))), self.blind_spots
        )
        self.max_bounds = reduce(
            (lambda tx, ty: tuple(max(x, y) for x, y in zip(tx, ty))), self.blind_spots
        )
        self.footprint_shape = tuple(
            ma - mi + 1 for mi, ma in zip(self.min_bounds, self.max_bounds)
        )

    def get_mask(self, blind_spot_for_channel, image_shape):
        """
        Get mask method

        Parameters
        ----------
        blind_spot_for_channel : tuple
        image_shape : tuple

        Returns
        -------
        Tuple of (mask, shape)

        """
        # First we prepare the mask:
        footprint = numpy.zeros(shape=self.footprint_shape, dtype=numpy.bool)
        coordinates = tuple(
            int(x - m) for x, m in zip(blind_spot_for_channel, self.min_bounds)
        )
        footprint[coordinates] = True
        shape = tuple(
            int(ceil(s / fs)) for fs, s in zip(self.footprint_shape, image_shape)
        )
        mask = numpy.tile(footprint, reps=shape)
        return mask, shape

    def transform(self, input_image, target_image):
        """
        Transform method

        Parameters
        ----------
        input_image : numpy.ndarray
        target_image : numpy.ndarray

        Returns
        -------
        Tuple of (transformed_input_image, transformed_target_image) or (transformed_input_image, None)

        """

        input_batch_list = []
        target_batch_list = []
        for blind_spot_for_batch in self.blind_spots:

            shift_for_roll = tuple(-x for x in blind_spot_for_batch)

            # prepare input image:
            shifted_input_image = numpy.roll(input_image, shift=shift_for_roll)
            input_channels_list = []
            for blind_spot_for_channel in self.blind_spots:

                mask, shape = self.get_mask(blind_spot_for_channel, input_image.shape)

                # Then, we extract values:
                input_channel = shifted_input_image[mask].reshape(shape)
                input_channels_list.append(input_channel)

            input_batch = numpy.stack(input_channels_list)
            input_batch_list.append(input_batch)

            # Prepare target image:
            if target_image is not None:
                shifted_target_image = numpy.roll(target_image, shift=shift_for_roll)
                mask, shape = self.get_mask(
                    (0,) * target_image.ndim, shifted_target_image.shape
                )

                # Then, we extract values:
                target_channel = shifted_target_image[mask].reshape(shape)
                target_batch = target_channel[numpy.newaxis, ...]
                target_batch_list.append(target_batch)

        transformed_input_image = numpy.stack(input_batch_list)
        if target_image is not None:
            transformed_target_image = numpy.stack(target_batch_list)

            # We keep the target image shape and dtype handy for the inverse transform:
            self.target_image_shape = target_image.shape
            self.target_image_dtype = target_image.dtype

            return transformed_input_image, transformed_target_image

        else:
            return transformed_input_image, None

    def inverse_transform(self, translated_transformed_image):
        """
        Inverse transform method

        Parameters
        ----------
        translated_transformed_image : numpy.ndarray

        Returns
        -------
        translated_image : numpy.ndarray

        """

        translated_image = numpy.empty(
            shape=self.target_image_shape, dtype=self.target_image_dtype
        )

        for batch_image, blind_spot in zip(
            translated_transformed_image, self.blind_spots
        ):
            mask, shape = self.get_mask(blind_spot, translated_image.shape)
            translated_image[mask] = batch_image.ravel()

        return translated_image



# flake8: noqa
"""
Grayscale BM3D denoising demo file, based on Y. Mäkinen, L. Azzari, A. Foi, 2019.
Exact Transform-Domain Noise Variance for Collaborative Filtering of Stationary Correlated Noise.
In IEEE International Conference on Image Processing (ICIP), pp. 185-189
"""
import numpy
import numpy as np
import matplotlib.pyplot as plt
from bm3d import bm3d

from aydin.features.fast.fast_features import FastFeatureGenerator
from aydin.io.datasets import camera, cropped_newyork, newyork, characters, pollen
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.cb import CBRegressor
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.bm3d.experiment_funcs import (
    get_experiment_noise,
    get_psnr,
    get_cropped_psnr,
)


def main():

    # Load noise-free image
    y = np.array(characters()) / 255
    # Possible noise types to be generated 'gw', 'g1', 'g2', 'g3', 'g4', 'g1w',
    # 'g2w', 'g3w', 'g4w'.
    noise_type = 'gw'
    noise_var = 0.2  # Noise variance
    seed = 0  # seed for pseudorandom noise realization

    # Generate noise with given PSD
    noise, psd, kernel = get_experiment_noise(noise_type, noise_var, seed, y.shape)
    # N.B.: For the sake of simulating a more realistic acquisition scenario,
    # the generated noise is *not* circulant. Therefore there is a slight
    # discrepancy between PSD and the actual PSD computed from infinitely many
    # realizations of this noise with different seeds.

    # Generate noisy image corrupted by additive spatially correlated noise
    # with noise power spectrum PSD
    z = np.atleast_3d(y) + np.atleast_3d(noise)

    # Call BM3D With the default settings.
    y_est = bm3d(z, psd)

    # To include refiltering:
    # y_est = bm3d(z, psd, 'refilter')

    # For other settings, use BM3DProfile.
    # profile = BM3DProfile(); # equivalent to profile = BM3DProfile('np');
    # profile.gamma = 6;  # redefine value of gamma parameter
    # y_est = bm3d(z, psd, profile);

    # Note: For white noise, you may instead of the PSD
    # also pass a standard deviation
    # y_est = bm3d(z, sqrt(noise_var));

    # Ignore values outside range for display (or plt gives an error for multichannel input)
    y_est = np.minimum(np.maximum(y_est, 0), 1)
    z_rang = np.minimum(np.maximum(z, 0), 1)
    plt.title("y, z, y_est")
    plt.imshow(np.concatenate((y, np.squeeze(z_rang), y_est), axis=1), cmap='gray')
    plt.show()

    generator = FastFeatureGenerator(include_spatial_features=False)

    regressor = LGBMRegressor(patience=8)  # , gpu=True

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    z = z.squeeze()
    it.train(z, z)
    denoised = it.translate(z)

    y = numpy.clip(y, 0, 1)
    z = numpy.clip(z, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    from skimage.metrics import peak_signal_noise_ratio as psnr
    from skimage.metrics import structural_similarity as ssim

    psnr = get_psnr(y, y_est)
    ssim_value = ssim(y, y_est)
    print("BM3D PSNR:", psnr)
    print("BM3D SSIM:", ssim_value)

    psnr = get_psnr(y, denoised)
    ssim_value = ssim(y, denoised)
    print("AYDIN PSNR:", psnr)
    print("AYDIN SSIM:", ssim_value)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(y, name='y')
        viewer.add_image(z, name='z')
        viewer.add_image(y_est, name='y_est')
        viewer.add_image(denoised, name='denoised')


if __name__ == '__main__':
    main()



import numpy

# from numpy.typing import ArrayLike
from skimage.exposure import equalize_adapthist, cumulative_distribution

from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class HistogramEqualisationTransform(ImageTransformBase):
    """Histogram Equalisation

    For some images with extremely unbalanced histograms, applying histogram
    equalisation will improve results. Two modes are supported: 'equalize',
    and 'clahe'.
    """

    preprocess_description = (
        "Apply histogram equalisation" + ImageTransformBase.preprocess_description
    )
    postprocess_description = (
        "Undo histogram equalisation" + ImageTransformBase.postprocess_description
    )
    postprocess_supported = True
    postprocess_recommended = True

    def __init__(
        self,
        mode: str = 'equalize',
        scale: float = 1.0 / 8,
        priority: float = 0.12,
        **kwargs,
    ):

        """
        Constructs a Histogram Transform

        Parameters
        ----------
        mode : str
            Two modes are supported: 'equalize', and 'clahe'.
        scale : float
            Scale of the kernel expressed relatively to the
            size of the image, values are within [0,1].
        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)
        self.mode = mode
        self.scale = scale
        self._cdf = None
        self._bin_centers = None
        self._original_dtype = None

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_cdf']
        del state['_bin_centers']
        del state['_original_dtype']
        return state

    def __str__(self):
        return f'{type(self).__name__}' f' (mode={self.mode},' f' scale={self.scale})'

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):

        with lsection(
            f"Equalises histogram for array of shape: {array.shape} and dtype: {array.dtype}"
        ):

            self._original_dtype = array.dtype
            array = array.astype(numpy.float32, copy=False)

            if self.mode == 'equalize':
                self._cdf, self._bin_centers = cumulative_distribution(array)
                new_array = _interpolation(array, self._bin_centers, self._cdf)
            elif self.mode == 'clahe':
                kernel_size = tuple(s / self.scale for s in array.shape)
                new_array = equalize_adapthist(array, kernel_size=kernel_size)
            else:
                raise ValueError(
                    f"Unsupported mode for histogram transform: {self.mode}"
                )

            return new_array

    def postprocess(self, array):

        if not self.do_postprocess:
            return array

        with lsection(
            f"Undoing histogram equalisation for array of shape: {array.shape} and dtype: {array.dtype}"
        ):
            array = array.astype(numpy.float32, copy=False)

            if self.mode == 'equalize':
                new_array = _interpolation(array, self._cdf, self._bin_centers)
            elif self.mode == 'clahe':
                # Inverse not supported yet:
                new_array = array
            else:
                raise ValueError(
                    f"Unsupported mode for histogram transform: {self.mode}"
                )

            # cast back to original dtype:
            new_array = new_array.astype(self._original_dtype, copy=False)

            return new_array


def _interpolation(image, x, y):
    out = numpy.interp(image.flat, x, y)
    return out.reshape(image.shape)



from typing import Tuple, Sequence

from scipy.ndimage import shift

from aydin.features.groups.base import FeatureGroupBase
from aydin.util.log.log import lprint


class TranslationFeatures(FeatureGroupBase):
    """
    Translations Feature Group class
    """

    def __init__(self, translations: Sequence[Tuple[int, ...]]):
        super().__init__()
        self.translations = list(translations)
        self.image = None
        self.excluded_voxels: Sequence[Tuple[int, ...]] = []

        self.kwargs = None

    @property
    def receptive_field_radius(self) -> int:
        radius = max(max(abs(d) for d in t) for t in self.translations)
        return radius

    def num_features(self, ndim: int) -> int:
        return len(self.translations)

    def prepare(self, image, excluded_voxels=None, **kwargs):
        if excluded_voxels is None:
            excluded_voxels = []

        self.image = image
        self.excluded_voxels = excluded_voxels
        self.kwargs = kwargs

    def compute_feature(self, index: int, feature):
        translation = self.translations[index]
        lprint(
            f"translation feature: {index}, translation={translation}, exclude_center={self.excluded_voxels}"
        )

        for excluded_voxel in self.excluded_voxels:
            if all(ev == t for ev, t in zip(excluded_voxel, translation)):
                return

        shift(
            self.image,
            shift=list(translation),
            output=feature,
            order=0,
            mode='constant',
            cval=0.0,
            prefilter=False,
        )

    def finish(self):
        # Here we cleanup any resource alocated for the last feature computation:
        self.image = None
        self.excluded_voxels = None
        self.kwargs = None



from tensorflow.python.keras import Input, layers
from tensorflow.python.keras.layers import LeakyReLU, GaussianNoise, Dense
from tensorflow.python.keras.models import Model
from tensorflow.python.keras.regularizers import l1


def block(
    x,
    outputs=1,
    layers=1,
    layer_name=None,
    trainable=True,
    initialiser=None,
    weight_decay=0.0,
):

    for i in range(0, layers):

        x = Dense(
            outputs,
            name=layer_name + 'd1l' + str(i),
            trainable=trainable,
            use_bias=False,
            kernel_initializer='glorot_uniform' if initialiser is None else initialiser,
            bias_initializer='zeros',
            kernel_regularizer=l1(weight_decay),
            bias_regularizer=l1(weight_decay),
        )(x)
        # x = BatchNormalization(name=layer_name + 'bn1', center=True, scale=False)(x)
        x = LeakyReLU(name=layer_name + 'act1l' + str(i))(x)

    return x


def yinyang(feature_dim, depth=16):
    input_feature = Input(shape=(feature_dim,), name='input')

    width = feature_dim

    x = input_feature
    outputs = []
    outputs.append(x)

    for d in range(0, depth):
        if d != 0:
            u = layers.concatenate([input_feature, x])
        else:
            u = x
        x = block(u, outputs=width, layer_name=f'fc{d}')
        outputs.append(x)

    x = layers.add([y for y in outputs])

    x = Dense(1, name='fc_last')(x)

    model = Model(input_feature, x)

    return model


def feed_forward_width(feature_dim, width=None, depth=16):

    if not width:
        width = feature_dim

    initialiser = 'glorot_uniform'  # RandomNormal(stddev = 0.1)

    input = Input(shape=(feature_dim,), name='input')

    if width <= feature_dim:
        x = input
    else:
        x = Dense(
            width - feature_dim,
            name='fc_first',
            kernel_initializer=initialiser,
            trainable=True,
        )(input)
        x = layers.concatenate([input, x])

    outputs = []
    outputs.append(x)

    for d in range(0, depth):
        x = block(x, outputs=width, layer_name=f'fc{d}', initialiser=initialiser)
        outputs.append(x)

    x = layers.add([y for y in outputs])

    x = Dense(1, name='fc_last')(x)

    model = Model(input, x)

    return model


def feed_forward(feature_dim, weight_decay=0.0001, depth=16, noise=None):

    width = feature_dim

    input = Input(shape=(feature_dim,), name='input')

    if noise is None:
        x = input
    else:
        x = GaussianNoise(noise)(input)

    outputs = []
    outputs.append(x)

    for d in range(0, depth):
        x = block(x, outputs=width, layer_name=f'fc{d}', weight_decay=weight_decay)
        outputs.append(x)

    x = layers.add([y for y in outputs])

    x = Dense(1, name='fc_last')(x)

    model = Model(input, x)

    return model


def yinyang2(feature_dim, depth=8):
    input_feature = Input(shape=(feature_dim,), name='input')

    width = feature_dim

    x = input_feature
    outputs = []
    outputs.append(x)

    for d in range(0, depth):
        if d != 0:
            u = layers.add([input_feature, x])
        else:
            u = x
        x = block(u, outputs=width, layer_name=f'fc{d}')
        outputs.append(x)

    x = layers.add([y for y in outputs])

    x = Dense(1, name='fc_last')(x)

    model = Model(input_feature, x)

    return model


def back_feed(feature_dim, depth=8):
    input_feature = Input(shape=(feature_dim,), name='input')

    width = feature_dim

    x = input_feature

    for d in range(0, depth):
        if d != 0:
            u = layers.add([input_feature, x])
        else:
            u = x
        x = block(u, outputs=width, layer_name=f'fc{d}')

    x = Dense(1, name='fc_last')(x)

    model = Model(input_feature, x)

    return model



from typing import Tuple

import numpy

from aydin.io import imread
from aydin.io.io import mapped_tiff
from aydin.it.base import ImageTranslatorBase
from aydin.util.log.log import lprint, lsection


class TimelapseDenoiser:
    """TimeLapse denoiser
    This translator is specialised for scaling up denoising to very long timelapses that can't fit in memory.
    This makes particularly sense for very large 3D+t datasets.
    For each time point we collect features from time points before and after as well as long-range averages (or medians)
    This helps to get good temporal consistency. To protect against self-hallucinations and speed up denoising, we train
    one denoiser per time point. It is advised to turn on spatial features. For timelapses with very large 3D stacks, it is
    recomended to use tiling.

    Attributes
    ----------

    Notes
    -----

    """

    def __init__(
        self,
        translator: ImageTranslatorBase,
        fine_temporal_window: int = 1,
        coarse_temporal_window: int = 7,
        use_median=True,
    ):
        """Constructs a timelapse denoiser given an image translator and the following parameters:

        Parameters
        ----------
        translator : ImageTranslatorBase
        coarse_temporal_window : int
        fine_temporal_window : int
        use_median : bool

        """
        self.use_median = use_median
        self.coarse_temporal_window = coarse_temporal_window
        self.fine_temporal_window = fine_temporal_window
        self.translator = translator

    def denoise_paths(
        self,
        input_image_path: str,
        denoised_image_path: str,
        batch_dims: Tuple[int] = None,
        channel_dims: Tuple[int] = None,
        tile_size=None,
        interval=None,
    ):
        """Convenience method to denoise an image from a path

        Parameters
        ----------
        input_image_path : str
        denoised_image_path : str
        batch_dims : tuple
        channel_dims : tuple
        tile_size
        interval

        """
        input_image_array, _ = imread(input_image_path)

        with mapped_tiff(
            denoised_image_path, input_image_array.shape, input_image_array.dtype
        ) as denoised_image_array:
            self.denoise(
                input_image_array,
                denoised_image_array,
                batch_dims,
                channel_dims,
                tile_size=tile_size,
                interval=interval,
            )

    def denoise(
        self,
        input_image_array: numpy.ndarray,
        denoised_image_array: numpy.ndarray = None,
        batch_dims: Tuple[int] = None,
        channel_dims: Tuple[int] = None,
        tile_size=None,
        interval=None,
    ):
        """Denoises image.
        Ideally, the array should not be fully loaded in memory but should be a 'lazy-loading' array.

        Parameters
        ----------
        input_image_array : numpy.ndarray
            image to denoise
        denoised_image_array : numpy.ndarray
            denoised image
        batch_dims : tuple
            tuple specifying which are the batch dimensions
        channel_dims : tuple
            tuple specifying which are the channel dimensions
        """
        with lsection(
            f"Denoising image with dimensions {input_image_array.shape} along first dimension"
        ):
            num_time_points = input_image_array.shape[0]

            # set default batch_dim and channel_dim values:
            if batch_dims is None:
                batch_dims = (False,) * len(input_image_array.shape)
            if channel_dims is None:
                channel_dims = (False,) * len(input_image_array.shape)

            if denoised_image_array is None:
                denoised_image_array = numpy.empty_like(input_image_array)

            tp_begin = 0 if interval is None else interval[0]
            tp_end = num_time_points if interval is None else interval[1]
            for tpi in range(tp_begin, tp_end):
                self._denoise_single_timepoint(
                    tpi,
                    num_time_points,
                    input_image_array,
                    denoised_image_array,
                    batch_dims,
                    channel_dims,
                    tile_size,
                )

            return denoised_image_array

    def _denoise_single_timepoint(
        self,
        tpi,
        num_time_points,
        input_image_array,
        denoised_image_array,
        batch_dims,
        channel_dims,
        tile_size,
    ):
        """Private method to denoise single given timepoint.

        Parameters
        ----------
        tpi
        num_time_points
        input_image_array
        denoised_image_array
        batch_dims
        channel_dims
        tile_size

        """
        with lsection(
            f"Denoising time point: {tpi} of shape: {input_image_array[tpi].shape}"
        ):
            ftw = self.fine_temporal_window
            with lsection(
                f"Adding fine temporal feature channels for range: [-{ftw},{ftw}]"
            ):

                def get_relative_timpepoint(rel_index):
                    index = tpi + rel_index
                    index = min(num_time_points - 1, index)
                    index = max(0, index)
                    lprint(f'Fine features delta={rel_index} ')
                    return input_image_array[index]

                fine_window_list = [
                    get_relative_timpepoint(i) for i in range(-ftw, +ftw + 1)
                ]

            ctw = self.coarse_temporal_window
            with lsection(
                f"Added coarse temporal feature channels for range: [-{ctw},{ctw}]"
            ):

                def get_average(extent):

                    if extent > 0:
                        index = tpi + extent
                        index = min(num_time_points - 1, index)

                        if index <= tpi:
                            return numpy.zeros_like(input_image_array[0])

                        lprint(f'Slice: {[tpi + 1, index + 1]}')
                        average_stack = input_image_array[tpi + 1 : index + 1]
                        average_stack = average_stack.astype(numpy.float32)
                        if self.use_median:
                            average_stack = numpy.median(
                                average_stack, axis=0, keepdims=False
                            )
                        else:
                            average_stack = numpy.sum(
                                average_stack, axis=0, keepdims=False
                            )
                        average_stack /= abs(tpi - (index))

                    else:
                        index = tpi + extent
                        index = max(0, index)
                        if index >= tpi:
                            return numpy.zeros_like(input_image_array[0])

                        lprint(f'Slice: {[index, tpi - 1 + 1]}')
                        average_stack = input_image_array[index : tpi - 1 + 1]
                        average_stack = average_stack.astype(numpy.float32)
                        if self.use_median:
                            average_stack = numpy.median(
                                average_stack, axis=0, keepdims=False
                            )
                        else:
                            average_stack = numpy.sum(
                                average_stack, axis=0, keepdims=False
                            )
                        average_stack /= abs((tpi + 1) - index)

                    lprint(f'Coarse temporal feature extent={extent}')
                    return average_stack

                coarse_window_list_past = [
                    get_average(ftw * 2 ** i) for i in range(1, ctw + 1)
                ]
                coarse_window_list_future = [
                    get_average(-ftw * 2 ** i) for i in range(1, ctw + 1)
                ]

            input_array_for_tp = numpy.stack(
                fine_window_list + coarse_window_list_past + coarse_window_list_future
            )
            target_array_for_tp = input_image_array[tpi : tpi + 1]

            channel_dims_tp = (True,) + tuple(c for c in channel_dims[1:])
            force_jinv = (
                (False,) * ftw + (True,) + (False,) * ftw + (False,) * (2 * ctw)
            )
            self.translator._passthrough_channels = (False,) * (2 * ftw + 1) + (
                True,
            ) * (2 * ctw)

            lprint(f'Channel dims: {channel_dims_tp} ')
            lprint(f'Force J-invariance tuple: {force_jinv} ')
            lprint(f'Pass-through channels: {self.translator._passthrough_channels} ')

            self.translator.train(
                input_array_for_tp,
                target_array_for_tp,
                batch_axes=batch_dims,
                channel_axes=channel_dims_tp,
                jinv=force_jinv,
            )

            self.translator.translate(
                input_array_for_tp,
                denoised_image_array[tpi],
                batch_axes=batch_dims,
                channel_axes=channel_dims_tp,
                tile_size=tile_size,
            )



import numpy
from numba import jit
from skimage.metrics import structural_similarity


@jit(nopython=True, parallel=True)
def mean_squared_error(image0, image1):
    return numpy.mean((image0 - image1) ** 2)


@jit(nopython=True, parallel=True)
def mean_absolute_error(image_a, image_b):
    return numpy.mean(numpy.absolute(image_a - image_b))


@jit(nopython=True, parallel=True)
def lhalf_error(image_a, image_b):
    return numpy.mean(numpy.absolute(image_a - image_b) ** 0.5) ** 2


def structural_loss(image_a, image_b):
    return numpy.astype(
        1 - structural_similarity(image_a, image_b), dtype=numpy.float32
    )



from qtpy.QtWidgets import QWidget, QHBoxLayout, QLabel, QPushButton
from aydin.gui._qt.custom_widgets.qt_range_slider import QHRangeSlider


class QRangeSliderWithLabels(QWidget):
    def __init__(self, parent, label="N/A", size=100):
        super(QWidget, self).__init__(parent)
        self.parent = parent
        self.size = size

        # Slider1
        self.widget_layout = QHBoxLayout()

        self.slider_label = QLabel(label, self)
        self.widget_layout.addWidget(self.slider_label, 0)
        self.two_columns = QLabel(":", self)
        self.widget_layout.addWidget(self.two_columns)

        self.slider = QHRangeSlider(
            initial_values=(0, size), data_range=(0, size), collapsible=False
        )

        self.slider.setStep(1.0)
        self.slider.valuesChanged.connect(self.slider_value_changed)

        self.widget_layout.addWidget(self.slider, 80)

        self.lower_limit_label = QPushButton("0")
        self.lower_limit_label.setDisabled(True)
        self.widget_layout.addWidget(self.lower_limit_label, 5)

        self.upper_limit_label = QPushButton(str(size))
        self.upper_limit_label.setDisabled(True)
        self.widget_layout.addWidget(self.upper_limit_label, 5)

        self.select_all_button = QPushButton("Select All")
        self.select_all_button.clicked.connect(
            lambda: self.slider.setValues(self.slider.range())
        )
        self.widget_layout.addWidget(self.select_all_button)

        self.setLayout(self.widget_layout)

    @property
    def lower_cutoff(self):
        return int(self.lower_limit_label.text())

    @property
    def upper_cutoff(self):
        return int(self.upper_limit_label.text())

    def slider_value_changed(self, event):
        lower, upper = self.slider.values()

        if self.slider_label.text() in ["X", "Y"] and upper - lower < 64:
            self.slider.setValues((self.lower_cutoff, self.upper_cutoff))
            return

        self.parent.update_crop_label_layer()
        self.lower_limit_label.setText(str(int(lower)))
        self.upper_limit_label.setText(str(int(upper)))
        self.parent.update_summary()



# flake8: noqa
import numpy
from numba import threading_layer
from scipy.ndimage import uniform_filter

from aydin.features.groups.uniform import UniformFeatures
from aydin.io.datasets import examples_single
from aydin.util.log.log import lsection, Log


def demo_uniform(image_name: str, image, repeats=32):
    Log.enable_output = True

    uniform = UniformFeatures()

    size = 3

    # Warmup:
    with lsection(f"Numba {image_name} (r={repeats}) WARMUP!:"):
        numba_filtered_image = uniform._compute_uniform_filter(image, size=size)

    with lsection(f"Numba {image_name} (r={repeats}):"):
        for _ in range(repeats):
            numba_filtered_image = uniform._compute_uniform_filter(image, size=size)

    with lsection(f"Scipy {image_name} (r={repeats}):"):
        for _ in range(repeats):
            scipy_filtered_image = uniform_filter(image, size=size, mode="nearest")

    numpy.testing.assert_array_almost_equal(
        numba_filtered_image, scipy_filtered_image, decimal=3
    )

    print("Threading layer chosen: %s" % threading_layer())


# image = fibsem(full=True).astype(numpy.float32)
# demo_numba_uniform("fibsem 2D", image)

image = examples_single.royerlab_hcr.get_array().squeeze()[0, ..., :320]
demo_uniform("islet 3D", image, repeats=10)
#
# image = examples_single.hyman_hela.get_array().squeeze()
# demo_numba_uniform("hela 4D", image, repeats=1)



import napari
import numpy
from aydin.gui._qt.custom_widgets.readmoreless_label import QReadMoreLessLabel
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QWidget

from aydin.gui._qt.custom_widgets.horizontal_line_break_widget import (
    QHorizontalLineBreakWidget,
)
from aydin.gui._qt.custom_widgets.qt_slider_with_labels import QRangeSliderWithLabels
from aydin.util.misc.units import human_readable_byte_size


class BaseCroppingTab(QWidget):
    """Use the sliders to select a region of the image to crop."""

    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.parent = parent

        self._image = None
        self._metadata = None
        self._images = []
        self.crop_layer = None

        self.tab_layout = QVBoxLayout()
        self.tab_layout.setAlignment(Qt.AlignTop)

        self.explanation_layout = QHBoxLayout()
        self.explanation_text = QReadMoreLessLabel(self, self.__doc__)
        self.explanation_layout.addWidget(self.explanation_text, 90)
        self.tab_layout.addLayout(self.explanation_layout)

        # Horizontal Line Break
        self.tab_layout.addWidget(QHorizontalLineBreakWidget(self))

        self.viewer_and_sliders_layout = QHBoxLayout()
        self.viewer_and_sliders_layout.setAlignment(Qt.AlignTop)
        # Image
        self.viewer_layout = QVBoxLayout()
        self.viewer_layout.setAlignment(Qt.AlignTop)
        self.viewer = napari.Viewer(show=False)

        self.viewer_layout.addWidget(self.viewer.window.qt_viewer)
        self.viewer_layout.addWidget(self.viewer.window.qt_viewer.dims)
        self.viewer_and_sliders_layout.addLayout(self.viewer_layout, 50)

        self.cropping_selection_layout = QVBoxLayout()
        self.cropping_selection_layout.addWidget(QLabel("Cropping Selection"))
        self.cropping_selection_layout.setAlignment(Qt.AlignTop)
        self.x_crop_slider = QRangeSliderWithLabels(self, label="X")
        self.cropping_selection_layout.addWidget(self.x_crop_slider)
        self.y_crop_slider = QRangeSliderWithLabels(self, label="Y")
        self.cropping_selection_layout.addWidget(self.y_crop_slider)
        self.z_crop_slider = QRangeSliderWithLabels(self, label="Z")
        self.cropping_selection_layout.addWidget(self.z_crop_slider)
        self.t_crop_slider = QRangeSliderWithLabels(self, label="T")
        self.cropping_selection_layout.addWidget(self.t_crop_slider)
        self.summary_layout = QVBoxLayout()
        self.summary_numbers_layout = QVBoxLayout()
        self.summary_numbers_layout.setAlignment(Qt.AlignCenter)
        self.summary_layout.setAlignment(Qt.AlignCenter)
        self.summary_nbvoxels_label = QLabel("Total number of voxels: N/A", self)
        self.summary_numbers_layout.addWidget(self.summary_nbvoxels_label)
        self.summary_nbytes_label = QLabel("Total size in bytes: N/A", self)
        self.summary_numbers_layout.addWidget(self.summary_nbytes_label)
        self.summary_layout.addLayout(self.summary_numbers_layout)

        self.warning_label_layout = QHBoxLayout()
        self.warning_label_layout.setAlignment(Qt.AlignCenter)
        self.size_warning_label = QLabel("", self)
        self.warning_label_layout.addWidget(self.size_warning_label)
        self.summary_layout.addLayout(self.warning_label_layout)
        self.cropping_selection_layout.addLayout(self.summary_layout)
        self.viewer_and_sliders_layout.addLayout(self.cropping_selection_layout, 50)

        self.tab_layout.addLayout(self.viewer_and_sliders_layout)

        self.setLayout(self.tab_layout)

    @property
    def images(self):
        return self.cropped_image if len(self._images) == 1 else self._images

    @images.setter
    def images(self, images):
        if len(images) == 0:
            self._images = []
            self.clear_cropping_tab()
            self.parent.enable_disable_a_tab(self.__class__, False)
        elif len(images) == 1:
            self.clear_cropping_tab()
            self.parent.enable_disable_a_tab(self.__class__, True)
            self._images = images
            self._metadata = images[0][2]
            self.image = images[0][1]
        else:
            self._images = [image[1] for image in images]
            self.clear_cropping_tab()
            self.parent.enable_disable_a_tab(self.__class__, False)

    @property
    def image(self):
        """image property

        Returns
        -------
        Returns the image with downscaled and padded spatiotemporal axes

        """
        return self._image

    @image.setter
    def image(self, image):
        """image property setter.

        Parameters
        ----------
        image

        """
        self._image = image
        self.new_crop = numpy.zeros(
            (
                self._image.shape[self._metadata.axes.find("Y")],
                self._image.shape[self._metadata.axes.find("X")],
            ),
            dtype=numpy.uint8,
        )

        self.initialize_viewer()
        self.update_sliders()

    @property
    def crop_selection_slicing_object(self):
        slider_slice = [slice(None)] * len(self.image.shape)
        for idx, axis in enumerate(self._metadata.axes):
            if axis == "X":
                lower, upper = self.x_crop_slider.slider.values()
                slider_slice[idx] = slice(int(lower), int(upper) + 1, 1)
            elif axis == "Y":
                lower, upper = self.y_crop_slider.slider.values()
                slider_slice[idx] = slice(int(lower), int(upper) + 1, 1)
            elif axis == "Z":
                lower, upper = self.z_crop_slider.slider.values()
                slider_slice[idx] = slice(int(lower), int(upper) + 1, 1)
            elif axis == "T":
                lower, upper = self.t_crop_slider.slider.values()
                slider_slice[idx] = slice(int(lower), int(upper) + 1, 1)

        return tuple(slider_slice)

    @property
    def cropped_image(self):
        """Takes the image apply crop and returns the cropped image.

        Returns
        -------

        """
        image = self._image[self.crop_selection_slicing_object]

        return [image]

    @property
    def selection_array(self):
        self.new_crop[...] = 0
        self.new_crop[
            (
                self.crop_selection_slicing_object[self._metadata.axes.find("Y")],
                self.crop_selection_slicing_object[self._metadata.axes.find("X")],
            )
        ] = 100

        return self.new_crop

    def update_sliders(self):
        self.x_crop_slider.slider.setRange(
            (0, self.image.shape[self._metadata.axes.find("X")])
        )
        self.x_crop_slider.slider.setValues(
            (0, self.image.shape[self._metadata.axes.find("X")])
        )
        self.x_crop_slider.upper_limit_label.setText(
            str(self.image.shape[self._metadata.axes.find("X")])
        )

        self.y_crop_slider.slider.setRange(
            (0, self.image.shape[self._metadata.axes.find("Y")])
        )
        self.y_crop_slider.slider.setValues(
            (0, self.image.shape[self._metadata.axes.find("Y")])
        )
        self.y_crop_slider.upper_limit_label.setText(
            str(self.image.shape[self._metadata.axes.find("Y")])
        )

        if "Z" in self._metadata.axes:
            self.z_crop_slider.setHidden(False)
            self.z_crop_slider.slider.setRange(
                (0, self.image.shape[self._metadata.axes.find("Z")])
            )
            self.z_crop_slider.slider.setValues(
                (0, self.image.shape[self._metadata.axes.find("Z")])
            )
            self.z_crop_slider.upper_limit_label.setText(
                str(self.image.shape[self._metadata.axes.find("Z")])
            )
        else:
            self.z_crop_slider.setHidden(True)

        if "T" in self._metadata.axes:
            self.t_crop_slider.setHidden(False)
            self.t_crop_slider.slider.setRange(
                (0, self.image.shape[self._metadata.axes.find("T")])
            )
            self.t_crop_slider.slider.setValues(
                (0, self.image.shape[self._metadata.axes.find("T")])
            )
            self.t_crop_slider.upper_limit_label.setText(
                str(self.image.shape[self._metadata.axes.find("T")])
            )
        else:
            self.t_crop_slider.setHidden(True)

    def initialize_viewer(self):
        self.viewer.layers.clear()
        self.viewer.add_image(self._image)

        self.crop_layer = self.viewer.add_image(self.selection_array)
        self.crop_layer.opacity = 0.4
        self.crop_layer.colormap = "cyan"

        for slider_widget in self.viewer.window.qt_viewer.dims.slider_widgets:
            slider_widget.layout().itemAt(0).widget().setMinimumWidth(75)
            slider_widget.layout().itemAt(1).widget().setMaximumWidth(75)
            slider_widget.layout().itemAt(1).widget().setText("Play")
            slider_widget.layout().itemAt(3).widget().setMinimumWidth(75)
            slider_widget.layout().itemAt(5).widget().setMinimumWidth(75)

        self.update_summary()

    def update_crop_label_layer(self):
        self.crop_layer.data = self.selection_array

    def update_summary(self):
        self.summary_nbvoxels_label.setText(
            f"Total number of voxels: {numpy.prod(self.cropped_image[0].shape)}"
        )
        self.summary_nbytes_label.setText(
            f"Total size in bytes: {human_readable_byte_size(self.cropped_image[0].nbytes)}"
        )

    def clear_cropping_tab(self):
        self.viewer.layers.clear()

        self.summary_nbvoxels_label.setText("Total number of voxels: N/A")
        self.summary_nbytes_label.setText("Total size in bytes: N/A")

    def on_data_model_update(self):
        self.images = self.parent.data_model.images_to_denoise



import dask
import numpy

from aydin.it.normalisers.base import NormaliserBase
from aydin.util.log.log import lsection, lprint


class MinMaxNormaliser(NormaliserBase):
    """Min-Max Normaliser"""

    def __init__(self, **kwargs):
        """Constructs a normalisers"""
        super().__init__(**kwargs)

    def calibrate(self, array):
        """Method to calibrate

        Parameters
        ----------
        array : numpy.ndarray

        """
        with lsection("Calibrating array using minmax method"):
            self.original_dtype = array.dtype

            if hasattr(array, '__dask_keys__'):
                self.rmin = dask.array.min(array.flatten()).compute()
                self.rmax = dask.array.max(array.flatten()).compute()
            else:
                self.rmin = numpy.min(array)
                self.rmax = numpy.max(array)

            lprint(f"Range for normalisation: [{self.rmin}, {self.rmax}]")

            return self.rmin, self.rmax



from aydin.restoration.denoise.noise2selffgr import noise2self_fgr  # noqa: F401
from aydin.restoration.denoise.noise2selfcnn import noise2self_cnn  # noqa: F401
from aydin.restoration.denoise.classic import Classic  # noqa: F401
from aydin.restoration.denoise.classic import classic_denoise  # noqa: F401
from aydin.restoration.deconvolve.lr import lucyrichardson  # noqa: F401

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "not-installed"



import locale
import math
import sys
import time
from contextlib import contextmanager


class Log:
    """
    Custom Log class
    """

    # current_section = ''
    gui_callback = None
    gui_statusbar = None
    guiEnabled = True
    enable_output = False
    depth = 0
    max_depth = math.inf
    log_elapsed_time = True
    override_test_exclusion = False

    # Define special characters:
    __vl__ = '│'  # 'Vertical Line'
    __br__ = '├'  # 'Branch Right'
    __bd__ = '├╗'  # 'Branch Down'
    __tb__ = '┴'  # 'Terminate Branch'
    __la__ = '«'  # 'Left Arrow'

    #  Windows terminal is dumb. We can't use our fancy characters from Yesteryears, sad:
    if (
        locale.getpreferredencoding() == "US-ASCII"
        or locale.getpreferredencoding() == "cp1252"
    ):
        __vl__ = '|'
        __br__ = '|->'
        __bd__ = '|\ '  # noqa: W605
        __tb__ = '-'
        __la__ = '<<'

    def __init__(self):
        return

    @staticmethod
    def native_print(*args, sep=' ', end='\n', file=sys.__stdout__):
        if Log.enable_output:
            print(*args, sep=sep, end=end, file=file)

        if Log.guiEnabled and Log.gui_callback is not None:
            result = ""
            for arg in args:
                result += str(arg)
                result += sep

            result += end
            Log.gui_callback.emit(result)

            if Log.gui_statusbar is not None:
                Log.gui_statusbar.showMessage(result)

    def set_log_elapsed_time(log_elapsed_time: bool):
        Log.log_elapsed_time = log_elapsed_time

    def set_log_max_depth(max_depth: int):
        Log.max_depth = max(0, max_depth - 1)


def lprint(*args, sep=' ', end='\n'):
    """
    Log print

    Parameters
    ----------
    args : list
    sep : str
    end : str

    """
    if not Log.override_test_exclusion:
        for arg in sys.argv:
            if "test" in arg:
                return

    if Log.depth <= Log.max_depth:
        level = min(Log.max_depth, Log.depth)
        Log.native_print(Log.__vl__ * int(level) + Log.__br__ + ' ', end='')
        Log.native_print(*args, sep=sep, end=end)


@contextmanager
def lsection(section_header: str):
    """
    Log section

    Parameters
    ----------
    section_header : str

    """
    if not Log.override_test_exclusion:
        for arg in sys.argv:
            if "test" in arg:
                yield
                return

    if Log.depth + 1 <= Log.max_depth:
        Log.native_print(
            Log.__vl__ * Log.depth + Log.__bd__ + ' ' + section_header
        )  # ≡
    elif Log.depth + 1 == Log.max_depth + 1:
        Log.native_print(
            Log.__vl__ * Log.depth
            + Log.__br__
            + f'= {section_header} (log tree truncated here)'
        )

    Log.depth += 1

    start = time.time()
    exception = None
    try:
        yield
    except Exception as e:
        exception = e

    stop = time.time()

    Log.depth -= 1
    if Log.depth + 1 <= Log.max_depth:

        if Log.log_elapsed_time:
            elapsed = stop - start

            if elapsed < 0.001:
                Log.native_print(
                    Log.__vl__ * (Log.depth + 1)
                    + Log.__tb__
                    + Log.__la__
                    + f' {elapsed * 1000 * 1000:.2f} microseconds'
                )
            elif elapsed < 1:
                Log.native_print(
                    Log.__vl__ * (Log.depth + 1)
                    + Log.__tb__
                    + Log.__la__
                    + f' {elapsed * 1000:.2f} milliseconds'
                )
            elif elapsed < 60:
                Log.native_print(
                    Log.__vl__ * (Log.depth + 1)
                    + Log.__tb__
                    + Log.__la__
                    + f' {elapsed:.2f} seconds'
                )
            elif elapsed < 60 * 60:
                Log.native_print(
                    Log.__vl__ * (Log.depth + 1)
                    + Log.__tb__
                    + Log.__la__
                    + f' {elapsed / 60:.2f} minutes'
                )
            elif elapsed < 24 * 60 * 60:
                Log.native_print(
                    Log.__vl__ * (Log.depth + 1)
                    + Log.__tb__
                    + Log.__la__
                    + f' {elapsed / (60 * 60):.2f} hours'
                )

        Log.native_print(Log.__vl__ * (Log.depth + 1))

        if exception is not None:
            raise exception



import random
from os.path import join
import keras.models
import numpy
from tensorflow.python.eager.context import device

from aydin.io.folders import get_temp_folder
from aydin.it.base import ImageTranslatorBase
from aydin.nn.models.jinet import JINetModel
from aydin.nn.models.unet import UNetModel
from aydin.nn.util.callbacks import (
    EarlyStopping,
    ReduceLROnPlateau,
    StopCenterGradient2D,
    StopCenterGradient3D,
)
from aydin.nn.util.data_util import random_sample_patches
from aydin.nn.util.validation_generator import train_image_generator
from aydin.regression.nn_utils.callbacks import ModelCheckpoint
from aydin.util.log.log import lsection, lprint
from aydin.util.tf.device import get_best_device_name, available_device_memory


class ImageTranslatorCNN(ImageTranslatorBase):
    """
    Convolutional Neural Network (CNN) based Image Translator<br>
    """

    verbose = 0

    def __init__(
        self,
        training_architecture: str = 'random',
        model_architecture: str = 'jinet',
        batch_size: int = 32,
        nb_unet_levels: int = 3,
        batch_norm: str = "instance",
        activation: str = 'ReLU',
        patch_size: int = 64,
        total_num_patches: int = None,
        adoption_rate: float = 0.5,
        mask_size: int = 5,
        random_mask_ratio: float = 0.1,
        max_epochs: int = 30,
        patience: int = 4,
        learn_rate: float = 0.01,
        **kwargs,
    ):
        """

        Parameters
        ----------
        training_architecture : str
            'shiftconv' or 'checkerbox' or 'random' or 'checkran' architecture
        model_architecture : str
            'unet' or 'jinet'
        batch_size : int
            Batch size for training
        nb_unet_levels : int
            Number of layers
        batch_norm
            Type of batch normalization (e.g. batch, instance)
        activation
        patch_size : int
            Size for patch sample e.g. 64 for (64, 64) or (64, 64, 64)
        total_num_patches
            Total number of patches for training
        adoption_rate
            % of random patches will be used for training, the rest will be discarded
        mask_size
            Mask shape for masking architecture; int of the same size as the spatial dimension
        random_mask_ratio
            Probability of masked pixels in random masking approach
        max_epochs : int
            Maximum number of epochs allowed
        patience : int
            Patience for EarlyStop or ReducedLR to be triggered
        learn_rate : float
            Initial learn rate
        kwargs
            Meant to have only keyword arguments for super class constructor. Do NOT abuse.
        """
        super().__init__(**kwargs)
        self.model_architecture = model_architecture  # both
        self.batch_size = batch_size  # both
        self.batch_norm = batch_norm  # both
        self.activation_fun = activation  # both
        self.patch_size = patch_size  # both
        self.total_num_patches = total_num_patches  # both
        self.adoption_rate = adoption_rate  # both
        self.max_epochs = max_epochs  # both
        self.patience = patience  # both
        self.learn_rate = learn_rate  # both
        self.model = None  # a CNN model  # both
        self.infmodel = None  # inference model  # both
        self.EStop_patience = self.patience * 2  # both
        self.ReduceLR_patience = self.patience  # both
        self.checkpoint = None  # both
        self.input_dim = None  # both
        self.stop_fitting = False  # both
        self.validation_images = None  # both
        self.validation_markers = None  # both
        self._create_patches_for_validation = (
            False  # if false use pixels for validation  # both
        )

        self.mask_size = mask_size  # unet
        self.random_mask_ratio = random_mask_ratio  # unet
        self.nb_unet_levels = nb_unet_levels  # unet
        self.training_architecture = training_architecture  # unet

        with lsection("CNN image translator"):
            lprint("training architecture: ", self.training_architecture)
            lprint("number of layers: ", self.nb_unet_levels)
            lprint("batch norm: ", self.batch_norm)
            lprint("mask size: ", self.mask_size)
            lprint("max_epochs", self.max_epochs)
            lprint("verbose: ", self.verbose)

    @property
    def model_class(self):
        if self.model_architecture == "jinet":
            return JINetModel
        elif self.model_architecture == "unet":
            return UNetModel
        else:
            raise ValueError("Unknown model architecture")

    def save(self, path: str):
        """
        Saves a 'all-batteries-included' image translation model at a given path (folder).

        Parameters
        ----------
        path : str
            path to save to

        Returns
        -------

        """
        with lsection(f"Saving 'CNN' image translator to {path}"):
            frozen = super().save(path)
            self.save_cnn(path)
        return frozen

    def save_cnn(self, path: str):
        if self.model is not None:
            # serialize model to JSON:
            self.model.save(join(path, "tf_model"))
            self.infmodel.save(join(path, "tf_inf_model"))
        else:
            lprint("There is no model to save yet.")

    def __getstate__(self):
        state = self.__dict__.copy()
        # exclude fields below that should/cannot be saved properly:
        del state['early_stopping']
        del state['reduce_learning_rate']
        del state['checkpoint']
        del state['model']
        del state['loss_history']
        del state['infmodel']
        del state['validation_images']

        return state

    def _load_internals(self, path: str):
        with lsection(f"Loading 'cnn' image translator from {path}"):
            # load JSON and create model:
            self.model = keras.models.load_model(join(path, "tf_model"))
            self.infmodel = keras.models.load_model(join(path, "tf_inf_model"))

    def get_receptive_field_radius(self, nb_unet_levels, shiftconv=False):
        if shiftconv:
            rf = 7 if nb_unet_levels == 0 else 36 * 2 ** (nb_unet_levels - 1) - 6
        else:
            rf = 3 if nb_unet_levels == 0 else 18 * 2 ** (nb_unet_levels - 1) - 4
        return int(rf // 2)

    def stop_training(self):
        """Stops currently running training within the instance by turning the flag
        true for early stop callback.

        """
        self.stop_fitting = True

    def retrain(self, input_image, target_image, training_architecture=None):
        self.training_architecture = training_architecture
        self.train(input_image, target_image)

    def _train(
        self,
        input_image,
        target_image,
        train_valid_ratio=0.1,
        callback_period=3,
        jinv=False,
    ):
        with device(get_best_device_name()):
            # Reshape the input image
            input_image = numpy.moveaxis(input_image, 1, input_image.ndim - 1)
            if not self.self_supervised:
                target_image = numpy.moveaxis(target_image, 1, target_image.ndim - 1)

            self.spacetime_ndim = input_image.ndim - 2
            if self.spacetime_ndim not in [2, 3]:
                raise ValueError(
                    "Number of spacetime dimensions have to be either 2 or 3."
                )

            self.input_dim = input_image.shape[1:]

            # Compute patch size from batch size
            if self.patch_size is None:
                self.patch_size = (
                    self.get_receptive_field_radius(
                        self.nb_unet_levels,
                        shiftconv='shiftconv' == self.training_architecture,
                    )
                    * 2
                )

                self.patch_size = (
                    self.patch_size - self.patch_size % 2 ** self.nb_unet_levels
                )

                if self.patch_size < 2 ** self.nb_unet_levels:
                    raise ValueError(
                        'Number of layers is too large for given patch size.'
                    )

            lprint(f'Patch size: {self.patch_size}')
            # TODO: Do we need to have one if statement to automatically convert self.batch_size = 1 for shiftconv?
            if 'shiftconv' in self.training_architecture or (
                self.model_architecture == "jinet" and self.spacetime_ndim == 3
            ):
                self.batch_size = 1
                lprint(
                    'When patch_size is assigned under shiftconv architecture, batch_size is automatically set to 1.'
                )

            # Adjust patch_size for given input shape
            self.patch_size = [self.patch_size] * self.spacetime_ndim

            # Check patch_size for unet models
            if 'unet' in self.model_architecture:
                patch_size = numpy.array(self.patch_size)
                if (patch_size.max() / (2 ** self.nb_unet_levels) <= 0).any():
                    raise ValueError(
                        f'Tile size is too small. The largest dimension of tile size has to be >= {2 ** self.nb_unet_levels}.'
                    )
                if (patch_size[-2:] % 2 ** self.nb_unet_levels != 0).any():
                    raise ValueError(
                        f'Tile sizes on XY plane have to be multiple of 2^{self.nb_unet_levels}'
                    )

            # Check if the smallest dimension of input data >= patch_size
            if min(self.patch_size) > min(self.input_dim[:-1]):
                smallest_dim = min(self.input_dim[:-1])
                self.patch_size[numpy.argsort(self.input_dim[:-1])[0]] = (
                    smallest_dim // 2 * 2
                )

            # Determine total number of patches
            if self.total_num_patches is None:
                self.total_num_patches = min(
                    input_image.size / numpy.prod(self.patch_size), 10240
                )  # upper limit of num of patches
                self.total_num_patches = (
                    self.total_num_patches
                    - (self.total_num_patches % self.batch_size)
                    + self.batch_size
                )
            else:
                if self.total_num_patches < self.batch_size:
                    raise ValueError(
                        'total_num_patches has to be larger than batch_size.'
                    )
                self.total_num_patches = (
                    self.total_num_patches
                    - (self.total_num_patches % self.batch_size)
                    + self.batch_size
                )

            lprint("Available mem: ", available_device_memory())
            lprint(f"Batch size for training: {self.batch_size}")

            # Decide whether to use validation pixels or patches
            if 1024 > input_image.size / numpy.prod(self.patch_size):
                with lsection(
                    f'Validation data will be created by monitoring {train_valid_ratio} of the pixels in the input data.'
                ):
                    img_train, img_val, val_marker = train_image_generator(
                        input_image, p=train_valid_ratio
                    )
            else:
                with lsection(
                    f'Validation data will be created by monitoring {train_valid_ratio} of the patches/images in the input data.'
                ):
                    self._create_patches_for_validation = True

            # Tile input and target image
            if self.patch_size is not None:
                with lsection('Random patch sampling...'):
                    lprint(f'Total number of patches: {self.total_num_patches}')
                    input_patch_idx = random_sample_patches(
                        input_image,
                        self.patch_size,
                        self.total_num_patches,
                        self.adoption_rate,
                    )

                    self.total_num_patches = len(input_patch_idx)

                    img_train_patch = []

                    if self._create_patches_for_validation:
                        for i in input_patch_idx:
                            img_train_patch.append(input_image[i])
                        img_train = numpy.vstack(img_train_patch)
                    else:
                        img_val_patch = []
                        marker_patch = []
                        for i in input_patch_idx:
                            img_train_patch.append(img_train[i])
                            img_val_patch.append(img_val[i])
                            marker_patch.append(val_marker[i])
                        img_train = numpy.vstack(img_train_patch)
                        img_val = numpy.vstack(img_val_patch)
                        val_marker = numpy.vstack(marker_patch)
                        self.validation_images = img_val
                        self.validation_markers = val_marker

                    if not self.self_supervised:
                        target_patch = []
                        for i in input_patch_idx:
                            target_patch.append(target_image[i])
                        target_image = numpy.vstack(target_patch)
                    else:
                        target_image = img_train

            # Last check of input size espetially for shiftconv
            if 'shiftconv' == self.training_architecture and self.self_supervised:
                # TODO: Hirofumi what is going on the conditional below :D  <-- check input dim is compatible w/ shiftconv
                if (
                    numpy.mod(
                        img_train.shape[1:][:-1],
                        numpy.repeat(
                            2 ** self.nb_unet_levels, len(img_train.shape[1:][:-1])
                        ),
                    )
                    != 0
                ).any():
                    raise ValueError(
                        'Each dimension of the input image has to be a multiple of 2^nb_unet_levels for shiftconv.'
                    )
                lprint(
                    'Model will be generated for self-supervised learning with shift convolution scheme.'
                )
                if numpy.diff(img_train.shape[1:][:2]) != 0:
                    raise ValueError(
                        'Make sure the input image shape is cubic as shiftconv mode involves rotation.'
                    )
                if (
                    numpy.mod(
                        img_train.shape[1:][:-1],
                        numpy.repeat(
                            2 ** (self.nb_unet_levels - 1),
                            len(img_train.shape[1:][:-1]),
                        ),
                    )
                    != 0
                ).any():
                    raise ValueError(
                        'Each dimension of the input image has to be a multiple of '
                        '2^(nb_unet_levels-1) as shiftconv mode involvs pixel shift. '
                    )

            shiftconv = (
                'shiftconv' == self.training_architecture and self.self_supervised
            )

            self.model = self.model_class(
                img_train.shape[1:],
                spacetime_ndim=self.spacetime_ndim,
                mini_batch_size=self.batch_size,
                nb_unet_levels=self.nb_unet_levels,
                normalization=self.batch_norm,
                activation=self.activation_fun,
                supervised=not self.self_supervised,
                shiftconv=shiftconv,
                learning_rate=self.learn_rate,
            )

            with lsection('CNN model summary:'):
                lprint(f'Model architecture: {self.model_architecture}')
                lprint(f'Train scheme: {self.training_architecture}')
                if self.model_architecture == 'unet':
                    lprint(f'Number of layers: {self.nb_unet_levels}')
                lprint(
                    f'Number of parameters in the model: {self.model.count_params()}'
                )
                lprint(f'Batch normalization: {self.batch_norm}')
                lprint(f'Training input size: {img_train.shape[1:]}')

            # End of train function and beginning of _train from legacy implmentation
            input_image = img_train

            with lsection(
                f"Training image translator from image of shape {input_image.shape} to image of shape {target_image.shape}:"
            ):

                if 'jinet' in self.model_architecture:
                    self.EStop_patience = self.EStop_patience + 10
                    self.ReduceLR_patience = self.ReduceLR_patience + 20

                # Early stopping patience:
                lprint(f"Early stopping patience: {self.EStop_patience}")

                # Effective LR patience:
                lprint(f"Effective LR patience: {self.ReduceLR_patience}")
                lprint(f'Batch size: {self.batch_size}')

                # Here is the list of callbacks:
                callbacks = []

                # Early stopping callback:
                self.early_stopping = EarlyStopping(
                    self, patience=self.EStop_patience, restore_best_weights=True
                )

                # Reduce LR on plateau:
                self.reduce_learning_rate = ReduceLROnPlateau(
                    verbose=1, patience=self.ReduceLR_patience, min_lr=1e-8, min_delta=0
                )
                self.reduce_learning_rate1 = ReduceLROnPlateau(
                    verbose=1,
                    patience=self.ReduceLR_patience,
                    min_lr=self.learn_rate * 0.01,
                    min_delta=0,
                )

                if self.checkpoint is None:
                    self.model_file_path = join(
                        get_temp_folder(),
                        f"aydin_cnn_keras_model_file_{random.randint(0, 1e16)}.hdf5",
                    )
                    lprint(f"Model will be saved at: {self.model_file_path}")
                    self.checkpoint = ModelCheckpoint(
                        self.model_file_path, verbose=1, save_best_only=True
                    )
                    # Add callbacks to the list:
                    callbacks.append(self.checkpoint)
                    callbacks.append(self.early_stopping)
                    callbacks.append(self.reduce_learning_rate)
                    if 'checkran' in self.training_architecture:
                        callbacks.append(self.reduce_learning_rate1)

                    if self.blind_spots:
                        self.blind_spots.remove((0, 0))

                    if self.spacetime_ndim == 2:
                        stop_center_gradient = StopCenterGradient2D(self.blind_spots)
                    elif self.spacetime_ndim == 3:
                        stop_center_gradient = StopCenterGradient3D(self.blind_spots)

                    callbacks = (
                        callbacks + [stop_center_gradient]
                        if 'jinet' in self.model_architecture and self.self_supervised
                        else callbacks
                    )

                # Convert mask_size to tuple
                self.mask_size = (self.mask_size,) * (input_image.ndim - 2)

                lprint("Training now...")
                if 'jinet' in self.model_architecture:
                    self.loss_history = self.model.fit(
                        input_image=input_image,
                        target_image=target_image,
                        max_epochs=self.max_epochs,
                        callbacks=callbacks,
                        verbose=self.verbose,
                        batch_size=self.batch_size,
                        total_num_patches=self.total_num_patches,
                        img_val=self.validation_images,
                        create_patches_for_validation=self._create_patches_for_validation,
                        train_valid_ratio=train_valid_ratio,
                    )
                else:
                    self.loss_history = self.model.fit(
                        input_image=input_image,
                        target_image=target_image,
                        callbacks=callbacks,
                        train_valid_ratio=train_valid_ratio,
                        img_val=self.validation_images,
                        val_marker=self.validation_markers,
                        training_architecture=self.training_architecture,
                        create_patches_for_validation=self._create_patches_for_validation,
                        total_num_patches=self.total_num_patches,
                        batch_size=self.batch_size,
                        random_mask_ratio=self.random_mask_ratio,
                        patch_size=self.patch_size,
                        mask_size=self.mask_size,
                        verbose=self.verbose,
                        max_epochs=self.max_epochs,
                        ReduceLR_patience=self.ReduceLR_patience,
                        parent=self,
                    )

    def _translate(self, input_image, image_slice=None, whole_image_shape=None):
        with device(get_best_device_name()):
            # Change dimensions to (B, space, C)
            input_image = numpy.moveaxis(input_image, 1, input_image.ndim - 1)

            # Check if padding is needed to have dim size of multiple of 2 in all dimension
            reshaped_for_cube = False
            reshaped_for_model = False
            spatial_shape = numpy.array(input_image.shape[1:-1])
            if abs(numpy.diff(spatial_shape)).min() != 0:
                reshaped_for_cube = True
                input_shape_max = numpy.ones(spatial_shape.shape) * spatial_shape.max()
                pad_square = (input_shape_max - spatial_shape) / 2
                pad_width1 = (
                    [[0, 0]]
                    + [
                        [
                            numpy.ceil(pad_square[i]).astype(int),
                            numpy.floor(pad_square[i]).astype(int),
                        ]
                        for i in range(len(pad_square))
                    ]
                    + [[0, 0]]
                )
                input_image = numpy.pad(input_image, pad_width1, 'edge')
                spatial_shape = numpy.array(input_image.shape[1:-1])

            if not (spatial_shape % 2 ** self.nb_unet_levels == 0).all():
                reshaped_for_model = True
                pad_width0 = (
                    2 ** self.nb_unet_levels
                    - (spatial_shape % 2 ** self.nb_unet_levels)
                    # + pad_square
                ) / 2
                pad_width2 = (
                    [[0, 0]]
                    + [
                        [
                            numpy.ceil(pad_width0[i]).astype(int),
                            numpy.floor(pad_width0[i]).astype(int),
                        ]
                        for i in range(len(pad_width0))
                    ]
                    + [[0, 0]]
                )
                input_image = numpy.pad(input_image, pad_width2, 'edge')

            # Change the batch_size in split layer or input dimensions accordingly
            kwargs_for_infmodel = {
                'mini_batch_size': 1,
                'nb_unet_levels': self.nb_unet_levels,
                'normalization': self.batch_norm,
                'activation': self.activation_fun,
                'shiftconv': 'shiftconv' == self.training_architecture,
                'spacetime_ndim': self.spacetime_ndim,
            }

            if len(input_image.shape[1:-1]) == 2:
                kwargs_for_infmodel['input_layer_size'] = [
                    None,
                    None,
                    input_image.shape[-1],
                ]
            elif len(input_image.shape[1:-1]) == 3:
                kwargs_for_infmodel['input_layer_size'] = [
                    None,
                    None,
                    None,
                    input_image.shape[-1],
                ]

                if self.model_architecture == "unet":
                    kwargs_for_infmodel['original_zdim'] = self.patch_size[0]

            if (
                'random' in self.training_architecture
                or 'check' in self.training_architecture
            ):
                kwargs_for_infmodel['supervised'] = True
            else:
                kwargs_for_infmodel['supervised'] = not self.self_supervised

            if self.infmodel is None:
                self.infmodel = self.model_class(**kwargs_for_infmodel)
            self.infmodel.set_weights(self.model.get_weights())

        try:
            output_image = self.infmodel.predict(
                input_image, batch_size=self.batch_size, verbose=self.verbose
            )
        except Exception:
            output_image = self.infmodel.predict(
                input_image, batch_size=1, verbose=self.verbose
            )

        # TODO: AhmetCan refactor
        if reshaped_for_model:
            if len(spatial_shape) == 2:
                output_image = output_image[
                    :,
                    pad_width2[1][0] : -pad_width2[1][1] or None,
                    pad_width2[2][0] : -pad_width2[2][1] or None,
                    :,
                ]
            else:
                output_image = output_image[
                    :,
                    pad_width2[1][0] : -pad_width2[1][1] or None,
                    pad_width2[2][0] : -pad_width2[2][1] or None,
                    pad_width2[3][0] : -pad_width2[3][1] or None,
                    :,
                ]
        if reshaped_for_cube:
            if len(spatial_shape) == 2:
                output_image = output_image[
                    :,
                    pad_width1[1][0] : -pad_width1[1][1] or None,
                    pad_width1[2][0] : -pad_width1[2][1] or None,
                    :,
                ]
            else:
                output_image = output_image[
                    :,
                    pad_width1[1][0] : -pad_width1[1][1] or None,
                    pad_width1[2][0] : -pad_width1[2][1] or None,
                    pad_width1[3][0] : -pad_width1[3][1] or None,
                    :,
                ]

        output_image = numpy.moveaxis(output_image, output_image.ndim - 1, 1)
        return output_image



# flake8: noqa

from aydin.gui.gui import run
import traceback
import warnings
import sys

if __name__ == "__main__":

    def warn_with_traceback(message, category, filename, lineno, file=None, line=None):
        log = file if hasattr(file, 'write') else sys.stderr
        traceback.print_stack(file=log)
        log.write(warnings.formatwarning(message, category, filename, lineno, line))

    warnings.showwarning = warn_with_traceback

    warnings.simplefilter("always")

    run('ver')



import math
import random

import numpy
from skimage.transform import downscale_local_mean

from aydin.util.log.log import lprint, lsection


class DataHistogramBalancer:
    """Training data balancer : Avoids having an excess of one class (y values) in the training data"""

    def __init__(
        self,
        number_of_bins=1024,
        keep_ratio=1,
        balance=True,
        use_median=False,
        favour_bright_pixels=False,
    ):
        """Constructs DataHistogramBalancer.

        Parameters
        ----------
        number_of_bins : int
        keep_ratio : float
        balance : bool
        use_median : bool
        favour_bright_pixels : bool
        """

        self.favour_bright_pixels = favour_bright_pixels
        self.use_median = use_median
        self.balance = balance

        self.number_of_bins = number_of_bins
        self.keep_ratio = keep_ratio

        self.min_value = 0
        self.max_value = 1

    def calibrate(
        self, array, batch_length, percentile=0.000001, num_batches_to_sample=128 * 1024
    ):
        """Calibrates the histogram balancer

        Parameters
        ----------
        array
        batch_length
        percentile
        num_batches_to_sample

        """

        array = array.ravel()

        # Calibration is not needed if we don't balance
        if not self.balance:
            lprint("Balancer: no calibration needed ")
            self.total_kept_counter = 0
            return

        # Find lower and upper bounds:
        if percentile > 0:
            full_min_value = numpy.percentile(array, 100 * percentile)
            full_max_value = numpy.percentile(array, 100 - 100 * percentile)
        else:
            full_min_value = array.min()
            full_max_value = array.max()

        with lsection("Balancer: sampling to estimate min and max values for entries"):
            # refine with sampling
            values_list = []
            array_length = array.size
            for i in range(num_batches_to_sample):
                start = random.randint(0, array_length - 1 - batch_length)
                stop = start + batch_length

                batch = array[..., start:stop]
                if self.use_median:
                    intensity = numpy.median(batch)
                else:
                    intensity = numpy.mean(batch)

                values_list.append(intensity)

            batch_min_value = min(values_list)
            batch_max_value = max(values_list)

        # Range:
        minmax_range = batch_max_value - batch_min_value

        # We extend the range by 5% on both sides to account for possible errors of the range estimate:
        self.min_value = batch_min_value - 0.05 * minmax_range
        self.max_value = batch_max_value + 0.05 * minmax_range

        # Keep the batch min and max within the full min and max (can happen because of percentile...)
        batch_min_value = max(full_min_value, batch_min_value)
        batch_max_value = min(full_max_value, batch_max_value)

        lprint(
            f"Balancer: full data min and max: [{full_min_value}, {full_max_value}] "
        )
        lprint(f"Balancer: batch min and max: [{batch_min_value}, {batch_max_value}] ")
        lprint(
            f"Balancer: effective min and max: [{self.min_value}, {self.max_value}] "
        )

    def initialise(self, total_entries):
        """Initialises the histogram balancer

        Parameters
        ----------
        total_entries

        """
        self.histogram_kept = numpy.zeros(self.number_of_bins)
        self.histogram_all = numpy.zeros(self.number_of_bins)

        self.total_entries = total_entries

        max_density = int(
            math.ceil(self.keep_ratio * self.total_entries / self.number_of_bins)
        )

        if self.favour_bright_pixels:
            self.max_entries_per_bin = [
                2 * max_density * i / self.number_of_bins
                for i in range(self.number_of_bins)
            ]
        else:
            self.max_entries_per_bin = [max_density] * self.number_of_bins

        # In the case of no balancing, we just keep track of the kept entries:
        self.total_kept_counter = 0

    def add_entry(self, array):
        """Adds an entry

        Parameters
        ----------
        array

        Returns
        -------
        bool

        """

        array = array.ravel()

        # This ensures that we stay within the bounds of the 'keep-ratio'
        fill_ratio = self.total_kept_counter / self.total_entries
        if fill_ratio > self.keep_ratio:
            return False

        if self.balance:

            if self.use_median:
                intensity = numpy.median(array)
            else:
                intensity = numpy.mean(array)

            # Rescale to range [0,1] range:
            intensity = (intensity - self.min_value) / (self.max_value - self.min_value)
            intensity = min(1, max(0, intensity))

            index = int(self.number_of_bins * intensity)
            index = min(len(self.histogram_kept) - 1, index)

            self.histogram_all[index] += 1

            if self.histogram_kept[index] < self.max_entries_per_bin[index]:
                self.histogram_kept[index] += 1
                self.total_kept_counter += 1
                return True

            return False

        else:
            # we just decimate uniformly according to the keep-ratio without histogram balancing:
            if random.random() < self.keep_ratio:
                self.total_kept_counter += 1
                return True

            return False

    def get_histogram_kept_as_string(self):
        """Returns the kept part of histogram as a string

        Returns
        -------
        str

        """

        if not self.balance:
            return '│ -- no balancing -- │'
        else:
            histogram_all = downscale_local_mean(
                self.histogram_all, factors=(8,), cval=0
            )
            maxvalue = numpy.max(histogram_all)
            histogram = downscale_local_mean(self.histogram_kept, factors=(8,), cval=0)
            return (
                '│'
                + ''.join((_value_to_fill_char(x / maxvalue)) for x in histogram)
                + '│'
            )

    def get_histogram_all_as_string(self):
        """Returns the whole histogram as a string

        Returns
        -------
        str

        """

        if not self.balance:
            return '│ -- no balancing -- │'
        else:
            histogram = downscale_local_mean(self.histogram_all, factors=(8,), cval=0)
            maxvalue = numpy.max(histogram)
            return (
                '│'
                + ''.join((_value_to_fill_char(x / maxvalue)) for x in histogram)
                + '│'
            )

    def get_histogram_dropped_as_string(self):
        """Returns the dropped part of histogram as a string

        Returns
        -------
        str

        """

        if not self.balance:
            return '│ -- no balancing -- │'
        else:
            histogram_all = downscale_local_mean(
                self.histogram_all, factors=(8,), cval=0
            )
            maxvalue = numpy.max(histogram_all)
            histogram_dropped = self.histogram_all - self.histogram_kept
            histogram = downscale_local_mean(histogram_dropped, factors=(8,), cval=0)
            return (
                '│'
                + ''.join((_value_to_fill_char(x / maxvalue)) for x in histogram)
                + '│'
            )

    def total_kept(self):
        """Total kept"""
        return (
            int(self.histogram_kept.sum()) if self.balance else self.total_kept_counter
        )

    def percentage_kept(self):
        """Percentage kept"""
        if self.balance:
            return min(1, self.histogram_kept.sum() / self.total_entries)
        else:
            return self.total_kept_counter / self.total_entries


def _value_to_fill_char(x):
    if x <= 0.00:
        return ' '
    elif x <= 0.01:
        return '·'
    elif x <= 0.05:
        return '-'
    elif x <= 0.10:
        return '■'
    elif x <= 0.25:
        return '░'
    elif x <= 0.50:
        return '▒'
    elif x <= 0.75:
        return '▓'
    elif x > 0.75:
        return '█'



from aydin.util.log.log import lprint


class ProgressBar:
    """Progress Bar

    Parameters
    ----------
    total : int

    """

    def __init__(self, total=100):
        self.total = total

    def emit(self, val):
        """Member method to emit

        Parameters
        ----------
        val : int

        """
        lprint("ProgressBar: ", val)
        if val > self.total:
            lprint("ProgressBar updated with a number bigger than it is maximum!")
